/*
 * rst.c
 *
 *  Created on: 17.07.2015
 *      Author: AndreR
 */

#include "rst.h"

#include "stm32f407xx.h"
#include "util/init_hal.h"
#include "util/console.h"
#include "util/log.h"
#include "errors.h"
#include "wifi.h"
#include <string.h>

#include "protobuf-c/radio_protocol_wrapper.pb-c.h"

#define LOCAL_VEL_MAX_ACC_XY	10.0f
#define LOCAL_VEL_MAX_ACC_W		100.0f
#define LOCAL_VEL_MAX_JERK_XY	100.0f
#define LOCAL_VEL_MAX_JERK_W	1000.0f

RST rst = { .cfg = {
	.enabled = 0,
	.rate = 50,
	.port = 10010,
}, };

typedef struct PACKED _BasicKDInput
{
	uint8_t kickSpeed;		// [Q3.5 m/s]
	uint8_t kickFlags;		// mode, dev
	uint8_t dribblerSpeed;	// [rpm/100]
} BasicKDInput;

typedef struct PACKED _LocalVelInput
{
	int16_t xyw[3];

	uint8_t accMaxXY;
	uint8_t accMaxW;
	uint8_t jerkMaxXY;
	uint8_t jerkMaxW;

	BasicKDInput kd;
} LocalVelInput;

static void freeMem(void *allocator_data, void *pointer);
static void* allocMem(void *allocator_data, size_t size);

void TIM4_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	TIM4->SR = 0;

	chSysLockFromISR();
	chBSemSignalI(&rst.semRun);
	chSysUnlockFromISR();

	CH_IRQ_EPILOGUE();
}

void RSTInit()
{
	chBSemObjectInit(&rst.semRun, 1);

	rst.allocator.alloc = &allocMem;
	rst.allocator.free = &freeMem;
	rst.allocator.allocator_data = 0;

	FlashFSOpenOrCreate("rst/cfg", 0, &rst.cfg, sizeof(RSTConfig), &rst.pCfgFile);

	for(uint8_t i = 0; i < NRF24_MAX_BOTS; i++)
	{
		SystemMatchCtrl* pCtrl = &wifi.bots[i].matchCtrl;

		pCtrl->curPosition[0] = SYSTEM_MATCH_CTRL_UNUSED_FIELD;
		pCtrl->curPosition[1] = SYSTEM_MATCH_CTRL_UNUSED_FIELD;
		pCtrl->curPosition[2] = SYSTEM_MATCH_CTRL_UNUSED_FIELD;
		pCtrl->skillId = 2; // local vel input
		pCtrl->feedbackFreq = 50;
	}

	RCC->APB1ENR |= RCC_APB1ENR_TIM4EN;

	TIM4->CR1 = 0;
	TIM4->CR2 = 0;
	TIM4->SR = 0;
	TIM4->DIER = TIM_DIER_UIE;
	TIM4->PSC = 83; 	// clock is now 1MHz
	TIM4->ARR = 999;	// = 1ms

	RSTSetRate(rst.cfg.rate);

	NVICEnableIRQ(TIM4_IRQn, IRQL_RST_TIMER);

	TIM4->CR1 = TIM_CR1_CEN;
}

void RSTSetRate(uint8_t rate)
{
	if(rate < 10)
		rate = 10;

	rst.cfg.rate = rate;

	uint32_t arr = 1000000/((uint32_t)rate);
	TIM4->ARR = arr;
}

void RSTEnable(uint8_t enable)
{
	rst.cfg.enabled = enable;
}

void RSTSetPort(uint16_t port)
{
	rst.cfg.port = port;
}

void RSTSaveConfig()
{
	FlashFSWrite(&rst.pCfgFile, &rst.cfg, sizeof(RSTConfig));
}

static uint8_t floatTo5Q3(float in)
{
	float kickSpeed = in;

	uint32_t kickSpeedA = (uint32_t)kickSpeed;
	uint32_t kickSpeedB = ((((uint32_t)(kickSpeed * 1000)) % 1000) + 15) / 32;

	if (kickSpeed > 7.96875f)
	{
		kickSpeedA = 7;
		kickSpeedB = 31;
	}

	uint32_t result = (kickSpeedA << 5) | kickSpeedB;

	return result;
}

int16_t RSTInput(uint8_t* pData, uint16_t length)
{
	rst.allocUsed = 0;	// free vision buffer

	RadioProtocolWrapper* pPacket = 0;
	pPacket = radio_protocol_wrapper__unpack(&rst.allocator, length, pData);
	if(pPacket == 0)
		return ERROR_RST_PARSE;

	for(size_t i = 0; i < pPacket->n_command; i++)
	{
		RadioProtocolCommand* pCmd = pPacket->command[i];

		int32_t id = pCmd->robot_id;
		if(id > 11)
			continue;

//		ConsolePrint("RST Bot: %d\r\n", id);

		// yellow bot online?
		if(wifi.bots[id].online == 0)
		{
			//no, blue one online?
			if(wifi.bots[id+12].online)
				id += 12;
			else
				continue;	// no, continue
		}

		SystemMatchCtrl* pCtrl = &wifi.bots[id].matchCtrl;
		pCtrl->skillId = 2;

		LocalVelInput* pVelCmd = (LocalVelInput*)pCtrl->skillData;

		pVelCmd->xyw[0] = pCmd->velocity_y*-1000.0f;
		pVelCmd->xyw[1] = pCmd->velocity_x*1000.0f;
		pVelCmd->xyw[2] = pCmd->velocity_r*1000.0f;

		pVelCmd->accMaxXY = (4.0f / LOCAL_VEL_MAX_ACC_XY) * 255;
		pVelCmd->jerkMaxXY = (50.0f / LOCAL_VEL_MAX_JERK_XY) * 255;
		pVelCmd->accMaxW = (50.0f / LOCAL_VEL_MAX_ACC_W) * 255;
		pVelCmd->jerkMaxW = (500.0f / LOCAL_VEL_MAX_JERK_W) * 255;
//		pVelCmd->accMaxXY = 255;
//		pVelCmd->jerkMaxXY = 255;
//		pVelCmd->accMaxW = 255;
//		pVelCmd->jerkMaxW = 255;

		if(pVelCmd->xyw[0] != 0 || pVelCmd->xyw[1] != 0 || pVelCmd->xyw[2] != 0)
		{
			wifi.bots[id].lastRSTTime = chVTGetSystemTimeX();
		}

//		ConsolePrint("X: %hd, Y: %hd, W: %hd, t: %u\r\n", *pX, *pY, *pW, wifi.bots[id].lastRSTTime);

		float rpm = 0;

		if(pCmd->has_dribbler_spin)
		{
			rpm = pCmd->dribbler_spin*-20000.0f;
			if(rpm < 0)
				rpm = 0;
			if(rpm > 25500.0f)
				rpm = 25500.0f;
		}

		pCtrl->flags = SYSTEM_MATCH_CTRL_FLAGS_KICKER_AUTOCHG;

		pVelCmd->kd.dribblerSpeed = (uint8_t)(rpm/100.0f);
		pVelCmd->kd.kickFlags = 0; // disarm
		pVelCmd->kd.kickSpeed = 0; // no kick speed

		if(pCmd->has_flat_kick)
		{
			pVelCmd->kd.kickFlags |= (KICKER_MODE_ARM << 4) | KICKER_DEVICE_STRAIGHT;

//			ConsolePrint("Flat kick\r\n");


			pVelCmd->kd.kickSpeed = floatTo5Q3(pCmd->flat_kick);

//			ConsolePrint("duration: %hu\r\n", pCtrl->kickSpeed);

			wifi.bots[id].lastARMTime = chVTGetSystemTimeX();
		}
		else if(pCmd->has_chip_kick)
		{
			pVelCmd->kd.kickFlags |= (KICKER_MODE_ARM << 4) | KICKER_DEVICE_CHIP;

//			ConsolePrint("Chip kick\r\n");

			pVelCmd->kd.kickSpeed = floatTo5Q3(pCmd->chip_kick);

			wifi.bots[id].lastARMTime = chVTGetSystemTimeX();
		}
	}

	return 0;
}

void RSTTask(void* params)
{
	(void)params;
	int16_t result;

	chRegSetThreadName("RST");

	static uint8_t sendBuf[sizeof(SystemMatchCtrl)+sizeof(PacketHeader)];
	PacketHeader* pHeader = (PacketHeader*)sendBuf;
	SystemMatchCtrl* pCtrl = (SystemMatchCtrl*)(sendBuf+sizeof(PacketHeader));
	pHeader->cmd = CMD_SYSTEM_MATCH_CTRL;
	pHeader->section = SECTION_SYSTEM;

	while(1)
	{
		chBSemWait(&rst.semRun);

		if(!rst.cfg.enabled)
			continue;

		for(uint8_t i = 0; i < NRF24_MAX_BOTS; i++)
		{
			WifiRobot* pBot = &wifi.bots[i];

			if(!pBot->online)
				continue;

			if(chVTTimeElapsedSinceX(pBot->lastRSTTime) > MS2ST(500))
			{
				pBot->matchCtrl.skillId = 0;	// motors off
			}

			if(chVTTimeElapsedSinceX(pBot->lastARMTime) > MS2ST(4000))
			{
//				uint8_t seq = (pBot->matchCtrl.kickFlags & SYSTEM_MATCH_CTRL_KICK_FLAGS_SEQ_MASK) >> 4;
//				++seq;
//
//				pBot->matchCtrl.kickFlags = SYSTEM_MATCH_CTRL_KICK_FLAGS_AUTOCHG_MASK | (seq << 4) | (3 << 1); // disarm
			}

			memcpy(pCtrl, &pBot->matchCtrl, sizeof(SystemMatchCtrl));

			result = NRF24QueueSend(&pBot->queue, sendBuf, sizeof(PacketHeader)+sizeof(SystemMatchCtrl)-SYSTEM_MATCH_CTRL_USER_DATA_SIZE+sizeof(LocalVelInput));
			if(result)
			{
				LogErrorC("No memory in queue\r\n", i);
				continue;
			}
		}
	}
}

static void* allocMem(void *allocator_data, size_t size)
{
	(void)allocator_data;

	if(rst.allocUsed + size > RST_ALLOC_BUF_SIZE)
	{
		ConsolePrint("WARN: RST out of memory\r\n");
		return 0;
	}

	void* pMem = rst.pAllocBuf + rst.allocUsed;

	rst.allocUsed += size;

	return pMem;
}

static void freeMem(void *allocator_data, void *pointer)
{
	(void)allocator_data;
	(void)pointer;
}
