/*
 * cli.c
 *
 *  Created on: 19.11.2012
 *      Author: AndreR
 */

#include "cli.h"

#include "hal/led.h"
#include "hal/kicker.h"
#include "hal/motor.h"
#include "hal/usart1.h"

#include "util/console.h"
#include "util/boot.h"
#include "util/intercom.h"
#include "util/sys_time.h"
#include "util/flash.h"
#include "util/flash_fs.h"
#include "util/crc.h"
#include "util/network_print.h"
#include "util/log.h"

#include "supply.h"
#include "intercom_constants.h"
#include "kick.h"
#include "cpu_load.h"
#include "io.h"
#include "test.h"

#include <string.h>

CLI cli;

static void processCommand();

void CLIInit()
{
	chMtxObjectInit(&cli.procBufMutex);

	static uint8_t rxData[CONSOLE_RX_SIZE];
	static uint8_t txData[CONSOLE_TX_SIZE];

	ConsoleData conData;
	conData.pDataTx = txData;
	conData.txSize = CONSOLE_TX_SIZE;
	conData.pDataRx = rxData;
	conData.rxSize = CONSOLE_RX_SIZE;
	conData.printSize = CONSOLE_PRINT_SIZE;
	conData.txCallback = 0;
	conData.pUserData = 0;

	ConsoleInit(conData);

	ConsoleIntercomAdapterInit(&cli.adapter, INTERCOM_ADDRESS_KD, INTERCOM_ADDRESS_MAIN);
}

void CLITask(void* params)
{
	(void)params;

	chRegSetThreadName("CLI");

#ifdef ENV_BOOT
	ConsolePrint("--- Bootloader (KD) ---\r\n");
#else
	ConsolePrint("--- KD ---\r\n");
#endif

	while(1)
	{
		chThdSleepMilliseconds(20);

		if(ConsoleIsCmdReady())
		{
			processCommand();

			ConsoleCmdProcessed();
		}
	}
}

static void processCommand()
{
	uint16_t u16;
	uint32_t u32[3];
	uint16_t vref;
	int32_t i32;

	if(ConsoleCmpCmd("test"))
	{
		ConsolePrint("Number: %d\r\n", 43);
	}

	if(ConsoleCmpCmd("test2"))
	{
		ConsolePrint("0123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789\r\n");
	}

	if(ConsoleCmpCmd("tasks"))
	{
		CPULoadPrintTasks();
	}

	if(ConsoleCmpCmd("cpu"))
	{
		CPULoadPrintUsage();
	}

	if(ConsoleScanCmd("vmax %hu", &u16) == 1)
	{
		ConsolePrint("vMax: %hu\r\n", u16);

		KickSetFireConfig(u16, kick.config.cooldown, kick.config.kickVoltageLimit);
	}

	if(ConsoleScanCmd("chg %hu", &u16) == 1)
	{
		if(u16)
		{
			ConsolePrint("Enable charging\r\n");
			KickerEnableCharge(1);
		}
		else
		{
			ConsolePrint("Disable charging\r\n");
			KickerEnableCharge(0);
		}
	}

	if(ConsoleScanCmd("autochg %hu", &u16) == 1)
	{
		if(u16)
		{
			ConsolePrint("Enable autocharging\r\n");
			KickerEnableAutoRecharge(1);
		}
		else
		{
			ConsolePrint("Disable autocharging\r\n");
			KickerEnableAutoRecharge(0);
		}
	}

	if(ConsoleCmpCmd("autodischg"))
	{
		ConsolePrint("Enabling auto discharging\r\n");
		KickerAutoDischarge();
	}

	if(ConsoleScanCmd("kstress %hu", &u16) == 1)
	{
		ConsolePrint("Setting auto-fire stress test: %hu\r\n", u16);

		kicker.stress = ((float)u16)/1000.0f;
	}

	if(ConsoleScanCmd("dis1 %hu", &u16) == 1)
	{
		ConsolePrint("Fire1: %hu\r\n", u16);

		KickerFire1(((float)u16)/1000.0f);
	}

	if(ConsoleScanCmd("arm1 %hu", &u16) == 1)
	{
		ConsolePrint("Arm1: %hu\r\n", u16);

		KickArm(((float)u16)/1000.0f, KICKER_DEVICE_STRAIGHT);
	}

	if(ConsoleScanCmd("dis2 %hu", &u16) == 1)
	{
		ConsolePrint("Fire2: %hu\r\n", u16);

		KickerFire2(((float)u16)/1000.0f);
	}

	if(ConsoleScanCmd("arm2 %hu", &u16) == 1)
	{
		ConsolePrint("Arm2: %hu\r\n", u16);

		KickArm(((float)u16)/1000.0f, KICKER_DEVICE_CHIP);
	}

	if(ConsoleCmpCmd("boot"))
	{
		BootSetBootloaderSelected(1);
		BootReset();
	}

	if(ConsoleCmpCmd("kstat"))
	{
		ConsolePrint("vCap: %d, iChg: %d\r\n", (int32_t)kicker.vCap, (int32_t)kicker.iChg);
		ConsolePrint("State: %hu, Auto: %hu\r\n", (uint16_t)kicker.state, (uint16_t)kicker.autoCharge);
	}

	if(ConsoleCmpCmd("kadc"))
	{
		for(uint8_t i = 0; i < SPI1_ADC_SAMPLES*2; i++)
			ConsolePrint("%hu ", kicker.rxData[i]);

		ConsolePrint("\r\n");
	}

	if(ConsoleScanCmd("cooldown %u", &u32[0]) == 1)
	{
		KickSetFireConfig(kick.config.maxVoltage, u32[0], kick.config.kickVoltageLimit);

		ConsolePrint("Kicker Cooldown set to %ums\r\n", kick.config.cooldown);
	}

	if(ConsoleCmpCmd("iradc"))
	{
		uint16_t irData[NUM_IR_SAMPLES];
		memcpy(irData, supply.irData, sizeof(uint16_t)*NUM_IR_SAMPLES);
		ConsolePrint("IRADC: ");
		for(uint16_t i = 0; i < NUM_IR_SAMPLES; i++)
			ConsolePrint("%hu ", irData[i]);
		ConsolePrint("\r\n");
	}

	if(ConsoleCmpCmd("ir"))
	{
		ConsolePrint("IR: %hu/%hu\r\n", supply.irResult[0], supply.irResult[1]);
	}

	if(ConsoleCmpCmd("supply"))
	{
		ConsolePrint("Supply: %hu/%hu\r\n", (uint16_t)(supply.bat*1000), (uint16_t)(supply.shunt*1000));
	}

	if(ConsoleScanCmd("ignore %hu", &u16) == 1)
	{
		if(u16)
		{
			io.ignoreMainCmds = 1;
			ConsolePrint("Ingoring commands from main\r\n");
		}
		else
		{
			io.ignoreMainCmds = 0;
			ConsolePrint("Main cmds re-enabled\r\n");
		}
	}

	if(ConsoleCmpCmd("ctrl"))
	{
		ConsolePrint("Ctrl: %hu\r\n", (uint16_t)motor.ctrlMode);
	}

 	if(ConsoleCmpCmd("ctrl none"))
	{
 		MotorSetCtrlMode(MOTOR_CTRL_MODE_NONE);

		ConsolePrint("Controller disabled\r\n");
	}

	if(ConsoleCmpCmd("ctrl speed"))
	{
 		MotorSetCtrlMode(MOTOR_CTRL_MODE_PID_SPEED);

		ConsolePrint("Speed PID controller active\r\n");
	}

	if(ConsoleCmpCmd("com none"))
	{
		MotorSetMode(MOTOR_MODE_DISABLE);

		ConsolePrint("Commutation mode: NONE\r\n");
	}

	if(ConsoleCmpCmd("com block"))
	{
		MotorSetMode(MOTOR_MODE_BLOCK);

		ConsolePrint("Commutation mode: BLOCK\r\n");
	}

	if(ConsoleScanCmd("vol %hu", &vref) == 1)
	{
		ConsolePrint("Setting dribbler voltage to: %humV\r\n", vref);

		MotorSetVoltage(vref*0.001f);
	}

	if(ConsoleScanCmd("rpm %i", &i32) == 1)
	{
		ConsolePrint("Setting dribbler to %i RPM\r\n", i32);

		MotorSpeedSetVelocity((float)i32);
	}

	if(ConsoleCmpCmd("adc"))
	{
		for(uint8_t a = 0; a < 2; a++)
		{
			ConsolePrint("ADC%hhu: ", a);
			for(uint8_t i = 0; i < DRIVE_ADC_SAMPLES; i++)
			{
				ConsolePrint("%5hu ", pMotor->adc[a][i]);
			}
			ConsolePrint("\r\n");
		}
	}

	if(ConsoleCmpCmd("dstat"))
	{
		ConsolePrint("Avg: %dmA\r\n", (int32_t)(motor.rmsCurrent*1000.0f));
		ConsolePrint("Temp: %d\r\n", (int32_t)(motor.temp*1000.0f));
		ConsolePrint("RPM: %d\r\nOverload: %hu\r\n", (int32_t)pMotor->curVelocity, (uint16_t)motor.pidSpeed.overload);
		ConsolePrint("PID output: %d\r\n", (int32_t)(motor.pidSpeed.output*1e3f));
		ConsolePrint("Bias: %d\r\n", (int32_t)(motor.biasCurrent.value*1000.0f));
	}

	if(ConsoleCmpCmd("hall"))
	{
		ConsolePrint("Hall: %hu\r\n", pMotor->halPos);
	}

	if(ConsoleCmpCmd("drib"))
	{
		ConsolePrint("RPM: %d\r\nOverload: %hu\r\n", (int32_t)pMotor->curVelocity, (uint16_t)motor.pidSpeed.overload);
	}

	if(ConsoleScanCmd("pid %u %u %u", u32, u32+1, u32+2) == 3)
	{
		ConsolePrint("%u, %u, %u\r\n", u32[0], u32[1], u32[2]);

		MotorSpeedSetPIDParams(u32[0]*1e-3f, u32[1]*1e-3f, u32[2]*1e-3f);
	}

	uint16_t angle;
	if(ConsoleScanCmd("va %hu %hu", &vref, &angle) == 2)
	{
		float r = ((float)vref)/1000.0f;

		MotorSetVoltageAngle(r, angle);
	}

	if(ConsoleCmpCmd("test dyn"))
	{
		ConsolePrint("Motor dynamics test\r\n");

		TestMotorDynamicsResult res;
		TestMotorDynamics(&res);

		ConsolePrint("R: %d, k: %d\r\n",
				(int32_t)(res.R*1000.0f), (int32_t)(res.k*1000000.0f));
		ConsolePrint("Current: %dmA, Speed: %drad/s\r\n",
				(int32_t)(res.cur*1000.0f), (int32_t)res.speed);
	}

	if(ConsoleCmpCmd("test hall"))
	{
		ConsolePrint("Motor hall test\r\n");

		TestMotorHallResult res;
		TestMotorHall(&res);

		for(uint16_t i = 0; i < 6; i++)
		{
			ConsolePrint("%3hd - %3hd - %3hd\r\n", res.pos[i].fwd, res.pos[i].ref, res.pos[i].rev);
		}
	}

	if(ConsoleCmpCmd("test unit"))
	{
		ConsolePrint("Power unit test\r\n");

		TestMotorPowerUnitResult res;
		TestMotorPowerUnit(&res);

		for(uint16_t ph = 0; ph < 3; ph++)
		{
			ConsolePrint("%hu: Main: %dmA, Mot: %dmA\r\n",
					ph, (int32_t)(res.mainCur[ph]*1000.0f), (int32_t)(res.motCur[ph]*1000.0f));
		}

		ConsolePrint("R1: %d, R2: %d, R3: %d\r\n", (int32_t)(res.R[0]*1000.0f), (int32_t)(res.R[1]*1000.0f), (int32_t)(res.R[2]*1000.0f));
	}

	if(ConsoleCmpCmd("full test"))
	{
		TestMotorPowerUnitResult power;
		TestMotorHallResult hall;
		TestMotorDynamicsResult dyn;

		ConsolePrint("\r\n=== DRIBBLER ===\r\n");

		ConsolePrint("--- Power Unit Test ---\r\n");
		TestMotorPowerUnit(&power);
		for(uint16_t ph = 0; ph < 3; ph++)
		{
			ConsolePrint("%hu: Main: %dmA, Mot: %dmA\r\n",
					ph, (int32_t)(power.mainCur[ph]*1000.0f), (int32_t)(power.motCur[ph]*1000.0f));
		}
		ConsolePrint("R1: %d, R2: %d, R3: %d (expected: 200)\r\n", (int32_t)(power.R[0]*1000.0f), (int32_t)(power.R[1]*1000.0f), (int32_t)(power.R[2]*1000.0f));

		ConsolePrint("--- Motor Hall Test ---\r\n");
		TestMotorHall(&hall);
		for(uint16_t i = 0; i < 6; i++)
		{
			ConsolePrint("%3hd - %3hd - %3hd\r\n", hall.pos[i].fwd, hall.pos[i].ref, hall.pos[i].rev);
		}

		ConsolePrint("--- Motor Dynamics Test ---\r\n");
		TestMotorDynamics(&dyn);
		ConsolePrint("R: %d (403), k: %d (2540)\r\n",
				(int32_t)(dyn.R*1000.0f), (int32_t)(dyn.k*1000000.0f));
		ConsolePrint("Current: %dmA (~1200), Speed: %drad/s (~3150)\r\n",
				(int32_t)(dyn.cur*1000.0f), (int32_t)dyn.speed);
	}

	if(ConsoleCmpCmd("flash hex"))
	{
		for(uint32_t i = 0; i < flashFS.bytesDel+flashFS.bytesUsed+48; i += 8)
		{
			uint8_t* d = (uint8_t*)(flashFS.flashAddr+i);
			ConsolePrint("0x%04X: %02X %02X %02X %02X %02X %02X %02X %02X\r\n",
					i, d[0], d[1], d[2], d[3], d[4], d[5], d[6], d[7]);
		}
	}

	if(ConsoleCmpCmd("flash list"))
	{
		FlashFSListFiles();
	}

	if(ConsoleCmpCmd("flash erase"))
	{
		ConsolePrint("Erasing flash\r\n");

		FlashErase(flashFS.flashAddr, flashFS.flashAddr+flashFS.flashSize);
	}

	if(ConsoleCmpCmd("time"))
	{
		ConsolePrint("Time: %uus\r\n", SysTimeUSec());
	}

	if(ConsoleCmpCmd("crc32"))
	{
		uint8_t data[12];
		uint32_t crc;
		uint32_t check;

		for(uint8_t i = 0; i < 12; i++)
			data[i] = (i+1)*4;

		crc = CRC32CalcChecksum(data, 8);
		ConsolePrint("8: 0x%08X\r\n", crc);
		check = CRC32VerifyChecksum(data, 8, crc);
		ConsolePrint("8: 0x%08X\r\n", check);

		crc = CRC32CalcChecksum(data, 7);
		ConsolePrint("7: 0x%08X\r\n", crc);
		check = CRC32VerifyChecksum(data, 7, crc);
		ConsolePrint("7: 0x%08X\r\n", check);

		crc = CRC32CalcChecksum(data, 6);
		ConsolePrint("6: 0x%08X\r\n", crc);
		check = CRC32VerifyChecksum(data, 6, crc);
		ConsolePrint("6: 0x%08X\r\n", check);

		crc = CRC32CalcChecksum(data, 5);
		ConsolePrint("5: 0x%08X\r\n", crc);
		check = CRC32VerifyChecksum(data, 5, crc);
		ConsolePrint("5: 0x%08X\r\n", check);
	}

	if(ConsoleCmpCmd("netprint"))
	{
		NetworkPrint("Hello Network - KD\n");
	}

	if(ConsoleScanCmd("log %hu", &u16) == 1)
	{
		LogPrintAllRecent(u16);
	}

	if(ConsoleCmpCmd("usart"))
	{
		ConsolePrint("--- USART1 ---\r\n");
		ConsolePrint("RX payload/wire/usage: %u / %u / %u%%\r\n", usart1.stats.rxPayload, usart1.stats.rxWire, (uint32_t)(usart1.stats.rxUsage*100.0f));
		ConsolePrint("TX payload/wire/usage: %u / %u / %u%%\r\n", usart1.stats.txPayload, usart1.stats.txWire, (uint32_t)(usart1.stats.txUsage*100.0f));
	}
}
