/*
 * kick.c
 *
 *  Created on: 21.05.2014
 *      Author: AndreR
 */

#include "kick.h"
#include "commands.h"
#include "util/fixed.h"
#include "intercom1.h"
#include "intercom_constants.h"
#include "hal/kicker.h"
#include "hal/motor.h"
#include "supply.h"
#include "util/network_print.h"
#include "util/sys_time.h"
#include "struct_ids.h"

#include <math.h>

#define EVENT_BARRIER_INTERRUPTED 0

#define KICKER_STATE_STANDBY		0
#define KICKER_STATE_IR_ARMED		1
#define KICKER_STATE_IR_ARMED_TIME	2
#define KICKER_STATE_DISCHARGE		3

KickGlobal kick = {
	.config = {
		.maxVoltage = 150,
		.chgHysteresis = 5.0f,
		.cooldown = 100,
		.irThreshold = 0.07,
		.kickVoltageLimit = 0.9f,
		.kickOffset = -0.02435f,
		.kickFactor = 0.6527f,
		.kickMax = 10.0f,
		.chipOffset = 1.823f,
		.chipFactor = 0.8027f,
		.chipMax = 6.0f,
	} };

static const ConfigFileDesc kickFileDesc =
	{ SID_CFG_KICK, 7, "kick/cfg", 11, (ElementDesc[]) {
		{ FLOAT, "max_voltage", "V", "Max. Voltage" },
		{ FLOAT, "chg_hyst", "V", "Charge Hysteresis" },
		{ UINT32, "cooldown", "ms", "Cooldown" },
		{ FLOAT, "ir_threshold", "V", "IR Threshold" },
		{ FLOAT, "ir_ignore", "", "Ignore Kick Threshold" },
		{ FLOAT, "kick_offset", "", "Kick Conv. Offset" },
		{ FLOAT, "kick_factor", "", "Kick Conv. Factor" },
		{ FLOAT, "kick_max", "", "Kick Conv. Max" },
		{ FLOAT, "chip_offset", "", "Chip Conv. Offset" },
		{ FLOAT, "chip_factor", "", "Chip Conv. Factor" },
		{ FLOAT, "chip_max", "", "Chip Conv. Max" },
	} };

static void irCallback()
{
	if(supply.irResult[1] > supply.irResult[0])
		kick.irDiff = supply.irResult[1] - supply.irResult[0];
	else
		kick.irDiff = supply.irResult[0] - supply.irResult[1];

	float irDiffVolt = kick.irDiff*0.001f;

	if(irDiffVolt < kick.config.irThreshold)
	{
		kick.interrupted = 1;
		chSysLockFromISR();
		chMBPostI(&kick.eventQueue, EVENT_BARRIER_INTERRUPTED);
		chSysUnlockFromISR();

		if(chVTTimeElapsedSinceX(kick.lastBarrierClearTime) > S2ST(10))
		{
			kick.barrierDamaged = 1;
		}
	}
	else
	{
		kick.interrupted = 0;
		kick.lastBarrierClearTime = chVTGetSystemTimeX();
		kick.barrierDamaged = 0;
	}
}

static void cfgCallback(uint16_t cfgId)
{
	if(cfgId == SID_CFG_KICK)
	{
		kicker.maxVoltage = kick.config.maxVoltage;
		kicker.cooldown = kick.config.cooldown;
		kicker.hystVoltage = kick.config.chgHysteresis;
	}
}

void KickInit()
{
	kick.pConfigFile = ConfigOpenOrCreate(&kickFileDesc, &kick.config, sizeof(KickerInternalConfig), &cfgCallback, 0);

	kick.state = KICKER_STATE_STANDBY;

	supply.irDoneCb = &irCallback;

	chMBObjectInit(&kick.eventQueue, kick.eventQueueData, KICK_EVENT_QUEUE_SIZE);

	ConfigNotifyUpdate(kick.pConfigFile);
}

void KickSetIRConfig(uint16_t irThreshold)
{
	kick.config.irThreshold = irThreshold;

	ConfigNotifyUpdate(kick.pConfigFile);
}

void KickSetFireConfig(float maxCap, uint32_t cooldown, float kickLimit)
{
	if(maxCap > 190.0f)
		maxCap = 190.0f;

	kick.config.maxVoltage = maxCap;
	kick.config.cooldown = cooldown;
	kick.config.kickVoltageLimit = kickLimit;

	ConfigNotifyUpdate(kick.pConfigFile);
}

void KickSetChgConfig(float hyst)
{
	kick.config.chgHysteresis = hyst;

	ConfigNotifyUpdate(kick.pConfigFile);
}

void KickTask(void* params)
{
	(void)params;

	msg_t event;
	msg_t result;

	chRegSetThreadName("Kick");

	while(1)
	{
		result = chMBFetch(&kick.eventQueue, &event, MS2ST(10));
		if(result == MSG_OK)
		{
			switch(event)
			{
				case EVENT_BARRIER_INTERRUPTED:
				{
					if(kicker.vCap > kick.config.kickVoltageLimit*kick.config.maxVoltage)
					{
						if(kick.state == KICKER_STATE_IR_ARMED)
							KickFireSpeed(kick.armed.speed, kick.armed.device);
						if(kick.state == KICKER_STATE_IR_ARMED_TIME)
							KickFire(kick.armed.speed, kick.armed.device);
					}
				}
				break;
			}
		}
	}
}

// speed is in [m/s]
void KickFireSpeed(float speed, uint8_t device)
{
	if(kicker.vCap < kick.config.kickVoltageLimit*kick.config.maxVoltage)
		return;

	if(device == KICKER_DEVICE_STRAIGHT)
	{
		if(speed > kick.config.kickMax)
			speed = kick.config.kickMax;

		float durMs = kick.config.kickOffset + kick.config.kickFactor*speed;
		if(durMs < 0)
			return;

		if(durMs > 10)
			durMs = 10.0f;

		KickerFire1(durMs*1e-3f);
	}

	if(device == KICKER_DEVICE_CHIP)
	{
		if(speed > kick.config.chipMax)
			speed = kick.config.chipMax;

		float durMs = kick.config.chipOffset + kick.config.chipFactor*speed;
		if(durMs < 0)
			return;

		if(durMs > 10)
			durMs = 10.0f;

		KickerFire2(durMs*1e-3f);
	}

	kick.state = KICKER_STATE_STANDBY;
}

void KickFire(float duration, uint8_t device)
{
	if(kicker.vCap < kick.config.kickVoltageLimit*kick.config.maxVoltage)
		return;

	if(device == KICKER_DEVICE_STRAIGHT)
		KickerFire1(duration);

	if(device == KICKER_DEVICE_CHIP)
		KickerFire2(duration);

	kick.state = KICKER_STATE_STANDBY;
}

void KickArmSpeed(float speed, uint8_t device)
{
	kick.armed.device = device;
	kick.armed.speed = speed;
	kick.state = KICKER_STATE_IR_ARMED;
}

void KickArm(float duration, uint8_t device)
{
	kick.armed.device = device;
	kick.armed.speed = duration;
	kick.state = KICKER_STATE_IR_ARMED_TIME;
}

void KickDisarm()
{
	kick.state = KICKER_STATE_STANDBY;
}

void KickDischarge()
{
	KickerAutoDischarge();
	kick.state = KICKER_STATE_STANDBY;
}
