/*
 * network_impl.c
 *
 *  Created on: 01.07.2010
 *      Author: AndreR
 */

#include "stm32f407xx.h"
#include "constants.h"

#include "main/ctrl_motor.h"
#include "main/ctrl_tigga.h"
#include "main/skills.h"
#include "main/network.h"

#include "util/intercom.h"
#include "util/console.h"
#include "util/init_hal.h"
#include "util/intercom_reliable.h"
#include "util/prog_store.h"
#include "hal/buzzer.h"
#include "cli.h"
#include "config_main.h"

void DMA2_Stream0_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	SPISyncDmaRxInterrupt(&network.nrf24.spi);

	CH_IRQ_EPILOGUE();
}

void EXTI4_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	EXTI->PR = (1 << 4);

	NRF24ExIRQ(&network.nrf24);

	CH_IRQ_EPILOGUE();
}

void TIM7_IRQHandler()
{
	CH_IRQ_PROLOGUE();

	TIM7->SR = 0;
	TIM7->CR1 = 0;

	chSysLockFromISR();
	chBSemSignalI(&network.timerSem);
	chSysUnlockFromISR();

	CH_IRQ_EPILOGUE();
}

void NetworkImplStartTimeout(uint32_t microSeconds)
{
	TIM7->CNT = 0;
	TIM7->ARR = microSeconds;
	TIM7->CR1 = TIM_CR1_CEN/* | TIM_CR1_OPM*/;
}

static void networkIntercomCallback(const IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)pUserData;

	if(pHeader->command == IC_NETWORK_NORMAL)
		NetworkSendPacket((PacketHeader*)pData, pData+PACKET_HEADER_SIZE, dataLength-PACKET_HEADER_SIZE);

	if(pHeader->command == IC_NETWORK_RELIABLE)
		NetworkSendPacketReliable((PacketHeader*)pData, pData+PACKET_HEADER_SIZE, dataLength-PACKET_HEADER_SIZE);
}

void NetworkImplInit()
{
	// NRF24 IRQ line setup
	GPIOInitData gpioInit;
	gpioInit.mode = GPIO_MODE_EXTI;
	gpioInit.pupd = GPIO_PUPD_UP;
	gpioInit.extiTrigger = GPIO_EXTI_TRIG_FALLING;
	GPIOInit(GPIOC, GPIO_PIN_4, &gpioInit);

	NVICEnableIRQ(EXTI4_IRQn, IRQL_SPI1);

	// TIM7 setup
	RCC->APB1ENR |= RCC_APB1ENR_TIM7EN;

	TIM7->CR1 = 0;
	TIM7->CR2 = 0;
	TIM7->SR = 0;
	TIM7->DIER = TIM_DIER_UIE;
	TIM7->PSC = 83; 	// clock is now 1MHz
	TIM7->ARR = 1000;	// = 1ms

	NVICEnableIRQ(TIM7_IRQn, IRQL_NRF24_TIMER);

	// SPI1 init
	RCC->APB2ENR |= RCC_APB2ENR_SPI1EN;

	NVICEnableIRQ(DMA2_Stream0_IRQn, IRQL_SPI1);

	gpioInit.mode = GPIO_MODE_AF;
	gpioInit.alternate = 5;
	gpioInit.otype = GPIO_OTYPE_PUSH_PULL;
	gpioInit.ospeed = GPIO_OSPEED_25MHZ;
	gpioInit.pupd = GPIO_PUPD_UP;
	GPIOInit(GPIOA, GPIO_PIN_5 | GPIO_PIN_6 | GPIO_PIN_7, &gpioInit);

	// NRF24 setup
	NRF24ExData data;

	static uint8_t rxData[NRF24_RX_SIZE];
	static uint8_t txData[NRF24_TX_SIZE];

	NRF24QueueInit(&network.queue, txData, NRF24_TX_SIZE, rxData, NRF24_RX_SIZE);
	network.queue.id = network.config.botId;

	data.csPin = GPIO_PIN_2;
	data.pCSPort = GPIOB;
	data.spiData.pRegister = SPI1;
	data.spiData.pDma = DMA2;
	data.spiData.dmaChRx = 0;
	data.spiData.dmaChTx = 3;
	data.rx = 1;
	data.channel = network.config.channel;
	data.speed = network.config.speed;
	data.pCEPort = GPIOA;
	data.cePin = GPIO_PIN_4;

	NRF24ExInit(&network.nrf24, &data);
	network.nrf24.slave.prescaler = SPI_CR1_BRDIV8;

	IntercomClient client;
	client.pUserData = 0;
	client.rxCallback = &networkIntercomCallback;
	client.section = INTERCOM_SECTION_NETWORK;

	IntercomAddClient(client);
}

void NetworkImplWifiStatsUpdate(PresenterWifiStat* pStats)
{
	IntercomHeader iHeader;
	iHeader.address = INTERCOM_ADDRESS_MEDIA;
	iHeader.section = INTERCOM_SECTION_PRESENTER;
	iHeader.command = IC_PRESENTER_WIFI_STAT;

	IntercomSend(&iHeader, sizeof(PresenterWifiStat), pStats);
}

// this is the main network processing function
void NetworkImplProcessPacket(PacketHeader* pHeader, uint8_t* pBuf, uint16_t dataLength)
{
	switch(pHeader->section)
	{
		case SECTION_SYSTEM:
		{
			switch(pHeader->cmd)
			{
				case CMD_SYSTEM_CONSOLE_COMMAND:
				{
					SystemConsoleCommand* pCmd = (SystemConsoleCommand*)pBuf;

					if(pCmd->target == INTERCOM_ADDRESS_MAIN)
					{
						// process onboard
						uint8_t* pData;
						uint32_t length = dataLength-sizeof(SystemConsoleCommand);
						int16_t result = FifoLinReserve(&console.rxFifo, length, &pData);
						if(result)
							return;

						memcpy(pData, pBuf+sizeof(SystemConsoleCommand), length);

						FifoLinCommit(&console.rxFifo, length);
					}
					else
					{
						if(cli.intercomFwd.inUse == 0)
						{
							cli.intercomFwd.bufSize = dataLength-sizeof(SystemConsoleCommand);

							if(cli.intercomFwd.bufSize <= CLI_INTERCOM_BUF_SIZE)
							{
								cli.intercomFwd.inUse = 1;
								cli.intercomFwd.target = pCmd->target;

								memcpy(cli.intercomFwd.buf, pBuf+sizeof(SystemConsoleCommand), cli.intercomFwd.bufSize);
							}
						}
					}
				}
				break;
			}
		}
		break;

#ifdef ENV_RUN
		case SECTION_CONFIG:
		{
			ConfigNetworkInput(pHeader, pBuf, dataLength);
		}
		break;
#endif
		default: break;
	}
}
