/*
 * robot_impl.c
 *
 *  Created on: 20.07.2016
 *      Author: AndreR
 */

#include "main/robot.h"
#include "main/ctrl.h"
#include "util/log_file.h"
#include "hal/led.h"
#include "spi2.h"
#include "motors.h"
#include "intercom1.h"
#include "util/sys_time.h"
#include "main/network.h"
#include "hal/buzzer.h"

BotParameters botParams = {
	.driveTrain = {
		.motor2WheelRatio = (15.0f/50.0f),
		.wheel2MotorRatio = (50.0f/15.0f),
		.motor = {
			.Km = 0.0255f,
			.Kn = (374.0f*2.0f*PI/60.0f),
			.Ke = (60.0f/(374.0f*2.0f*PI)),
			.Kf = 0.94e-4f,
			.R = 1.2f,
			.L = 560e-6f,
		}
	},
	.physical = { 0.0245f, 30.0f, 45.0f, 0.076f, 3.0f },
	.v2016 = 0,
};

static KdFeedback lastKdFeedback;

static void kdCallback(const IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)dataLength;
	(void)pUserData;

	switch(pHeader->command)
	{
		case IC_KD_FEEDBACK:
		{
			KdFeedback* pFb = (KdFeedback*)pData;

			memcpy(&lastKdFeedback, pFb, sizeof(KdFeedback));
		}
		break;
	}
}

void RobotImplInit()
{
	// register KD callback
	IntercomClient client;
	client.pUserData = 0;
	client.rxCallback = &kdCallback;
	client.section = INTERCOM_SECTION_KD;
	IntercomAddClient(client);
}

void RobotImplUpdateSensors(RobotSensors* pSensors)
{
	// motors
	pSensors->enc.time = motors.mot[0].cur.velocityTimestamp;
	pSensors->motorVol.time = motors.mot[0].cur.setVoltageTimestamp;

	for(uint8_t m = 0; m < 4; m++)
	{
		pSensors->enc.vel[m] = motors.mot[m].cur.velocity * (2*PI);
		pSensors->motorVol.vol[m] = motors.mot[m].cur.setVoltage;
	}

	// power
	pSensors->power.voltage = lastKdFeedback.power.bat;
	pSensors->power.current = lastKdFeedback.power.cur;
	pSensors->power.batEmpty = lastKdFeedback.power.batEmpty;

	// IMU
	pSensors->acc.linAcc[0] = spi2.acc[0];
	pSensors->acc.linAcc[1] = spi2.acc[1];
	pSensors->acc.linAcc[2] = spi2.acc[2];
	pSensors->acc.time = spi2.lastAccTime;
	pSensors->acc.updated = 1;

	pSensors->gyr.rotVel[0] = spi2.gyr[0];
	pSensors->gyr.rotVel[1] = spi2.gyr[1];
	pSensors->gyr.rotVel[2] = spi2.gyr[2];
	pSensors->gyr.time = spi2.lastGyrTime;
	pSensors->gyr.updated = 1;

	// Vision - SKIPPED (filled by robot base implementation)

	// Kicker
	pSensors->kicker.level = lastKdFeedback.kicker.cap;
	pSensors->kicker.chgCurrent = lastKdFeedback.kicker.chg;
	pSensors->kicker.kickCounter = lastKdFeedback.kicker.kickCounter;
	pSensors->kicker.flags = 0;
	if(lastKdFeedback.kicker.ready)
		pSensors->kicker.flags |= KICKER_FLAGS_READY;

	if(lastKdFeedback.barrier.damaged)
		pSensors->kicker.flags |= KICKER_FLAGS_DMG_BARRIER;

	// Dribbler
	pSensors->dribbler.speed = lastKdFeedback.dribbler.velocity*2.0f/4.4f; // motor RPM to bar RPM
	pSensors->dribbler.temp = lastKdFeedback.dribbler.temp;
	pSensors->dribbler.overheated = lastKdFeedback.dribbler.overheated;

	// Barrier
	pSensors->ir.level[0] = lastKdFeedback.barrier.level[0]*0.001f;
	pSensors->ir.level[1] = lastKdFeedback.barrier.level[1]*0.001f;
	pSensors->ir.interrupted = lastKdFeedback.barrier.interrupted;
}

void RobotImplWriteLogMsg(void* pMsg)
{
	LogEntryHeader* pHeader = (LogEntryHeader*)pMsg;

	IntercomHeader iHeader;
	iHeader.address = INTERCOM_ADDRESS_MEDIA;
	iHeader.section = INTERCOM_SECTION_LOGFILE;
	iHeader.command = IC_LOGFILE_MESSAGE;

	IntercomSend(&iHeader, pHeader->length, pMsg);
}

void RobotImplMotorOutput(const RobotCtrlOutput* pMotor)
{
	motors.ctrlMode = pMotor->ctrlMode;

	for(uint8_t i = 0; i < 4; i++)
	{
		motors.mot[i].set.velocity = pMotor->motorVel[i] * 1/(2*M_PI);
		motors.mot[i].set.voltage = pMotor->motorVoltage[i];
	}
}

void RobotImplKickerOutput(const KickerInput* pKicker)
{
	KdFire fire = {pKicker->mode, pKicker->device, pKicker->speed};
	IntercomHeader iHeader = { { INTERCOM_ADDRESS_KD, INTERCOM_SECTION_KD, IC_KD_FIRE } };

	IntercomSend(&iHeader, sizeof(KdFire), &fire);
}

void RobotImplDribblerOutput(const DribblerInput* pDribbler)
{
	IntercomHeader iHeader = { { INTERCOM_ADDRESS_KD, INTERCOM_SECTION_KD, IC_KD_DRIBBLER } };

	IntercomSend(&iHeader, sizeof(DribblerInput), pDribbler);
}

void RobotImplBuzzerPlay(uint16_t seqId)
{
	BuzzerPlayId(seqId);
}

void RobotImplKickerAutoCharge(uint8_t enable)
{
	KdAutoCharge chg = {enable};
	IntercomHeader iHeader = { { INTERCOM_ADDRESS_KD, INTERCOM_SECTION_KD, IC_KD_AUTOCHG } };

	IntercomSend(&iHeader, sizeof(KdAutoCharge), &chg);
}

void RobotImplSetLEDs(uint8_t state)
{
	(void)state;

	// no LEDs available :(
}
