/*
 * test.c
 *
 *  Created on: 27.06.2016
 *      Author: AndreR
 */

#include "test.h"
#include "hal/motors.h"
#include "hal/led.h"
#include "power.h"
#include "main/network.h"
#include "arm_math.h"
#include "util/arm_mat_util_f32.h"
#include "gui/test_drive.h"
#include "util/console.h"
#include "main/robot.h"

void printTestMotorDynamicsResult(TestMotorDynamicsResult* pResult);
void calculateReasoningOnTestResultsDrive(const TestMotorDynamicsResult* pResult, TestDynamicsReasoningResult* pReasoningResult);
void calculateReasoningOnTestResultsDribbler(const TestMotorDynamicsResult* pResult, TestDynamicsReasoningResult* pReasoningResult);
void calculateReasoningOnTestResults(const TestMotorDynamicsResult* pResult, TestDynamicsReasoningResult* pReasoningResult);
void printReasoningResultConstant(uint8_t result);
void printTestMotorDynamicsResult(TestMotorDynamicsResult* pResult);


TestData test;

void TestInit()
{
	chMBObjectInit(&test.eventQueue, test.eventQueueData, TEST_EVENT_QUEUE_SIZE);

	test.pDriveProg = &TestDriveProgress;
}

void TestTask(void* params)
{
	(void)params;

	msg_t event;

	chRegSetThreadName("Test");

	while(1)
	{
		if(chMBFetch(&test.eventQueue, &event, TIME_INFINITE) != MSG_OK)
			continue;

		uint8_t testId = event & 0xFF;
		uint8_t motorId = event >> 8;

		switch(testId)
		{
			case IC_TEST_MOT_DYN:
			{
				TestMotorDynamicsResult result;
				TestMotorDynamics(motorId, &result);

				TestDriveDynamicResult(&result);
			}
			break;
		}
	}
}

void TestScheduleMotorTest(uint8_t testId, uint8_t motorId)
{
	msg_t event = motorId;
	event <<= 8;
	event |= testId;

	chMBPost(&test.eventQueue, event, TIME_IMMEDIATE);
}

#define EST_SIZE 12
void TestMotorDynamics(uint8_t motorId, TestMotorDynamicsResult* pResult)
{
	EMAFilter current;
	EMAFilter speed;

	arm_matrix_instance_f32 A = {EST_SIZE, 2, (float[EST_SIZE*2]){}};
	arm_matrix_instance_f32 Ai = {2, EST_SIZE, (float[EST_SIZE*2]){}};
	arm_matrix_instance_f32 b = {EST_SIZE, 1, (float[EST_SIZE*1]){}};
	arm_matrix_instance_f32 x = {2, 1, (float[2]){}};

	TestProgress prog;
	prog.testId = IC_TEST_MOT_DYN;
	prog.motorId = motorId;
	prog.percentComplete = 0;

	if(test.pDriveProg)
		(*test.pDriveProg)(&prog);

	pResult->motorId = motorId;

	RobotTestModeEnable(1);
	RobotTestModeStopAll();

	chThdSleepMilliseconds(5);

	arm_mat_zero_f32(&A);
	arm_mat_zero_f32(&Ai);
	arm_mat_zero_f32(&b);
	arm_mat_zero_f32(&x);

	EMAFilterInit(&current, 0.9f);
	EMAFilterInit(&speed, 0.95f);

	for(uint8_t i = 0; i < 4; i++)
		MotorsDriveSetVoltage(i, 0);

	MotorsDribblerSetVoltage(0);

	for(uint16_t j = 0; j < 1000; j++)
	{
		EMAFilterUpdate(&current, power.iCur);
		chThdSleepMilliseconds(1);
	}

	float mainOffset = current.value;

	prog.percentComplete = 1/18.0f*100.0f;
	if(test.pDriveProg)
		(*test.pDriveProg)(&prog);

//	ConsolePrint("Main current offset: %d\r\n", (int32_t)(mainOffset*1000.0f));

	for(int16_t i = 0; i <= EST_SIZE; i++)
	{
		float volt = i-EST_SIZE/2;

		if(i == EST_SIZE/2)
			continue;

		if(motorId > 3)
			MotorsDribblerSetVoltage(fabsf(volt));
		else
			MotorsDriveSetVoltage(motorId, volt);

		chThdSleepMilliseconds(200);

		for(uint16_t j = 0; j < 800; j++)
		{
			EMAFilterUpdate(&current, power.iCur-mainOffset);

			if(motorId > 3)
				EMAFilterUpdate(&speed, motors.dribbler.velocity);
			else
				EMAFilterUpdate(&speed, motors.drive[motorId].velocity);

			chThdSleepMilliseconds(1);
		}

//		float mainCur = pDrive->bat*mainCurrent.value/volt;

		if(volt < 0)
			MAT_ELEMENT(A, i-1, 0) = -current.value;
		else
			MAT_ELEMENT(A, i-1, 0) = current.value;

		if(volt < 0 && motorId > 3)
			MAT_ELEMENT(A, i-1, 1) = -speed.value;
		else
			MAT_ELEMENT(A, i-1, 1) = speed.value;
		MAT_ELEMENT(b, i-1, 0) = volt;

		prog.percentComplete = ((float)(i+2))/18.0f*100.0f;
		if(test.pDriveProg)
			(*test.pDriveProg)(&prog);

//		ConsolePrint("U: %dmV, I: %dmA, v: %dmrad/s\r\n",
//				(int32_t)(volt*1000.0f), (int32_t)(current.value*1000.0f), (int32_t)(speed.value*1000.0f));
	}

	if(motorId > 3)
		MotorsDribblerSetVoltage(0);
	else
		MotorsDriveSetVoltage(motorId, 0);

	arm_mat_pinv(&A, &Ai);
	arm_mat_mult_f32(&Ai, &b, &x);

	float R = MAT_ELEMENT(x, 0, 0);
	float k = MAT_ELEMENT(x, 1, 0);

	pResult->R = R;
	pResult->k = k;
	pResult->cur = current.value;
	pResult->speed = speed.value;

	prog.percentComplete = 100;
	if(test.pDriveProg)
		(*test.pDriveProg)(&prog);

	RobotTestModeEnable(0);

	chThdSleepMilliseconds(1000);

	printTestMotorDynamicsResult(pResult);
}

void calculateReasoningOnTestResultsDrive(const TestMotorDynamicsResult* pResult, TestDynamicsReasoningResult* pReasoningResult) {
	if(pResult->R < 0.3f || pResult->R > 0.6f)
	{
		pReasoningResult->res = TEST_REASONING_RESULT_WARNING;
	}

	if(pResult->R < 0.2f || pResult->R > 0.7f)
	{
		pReasoningResult->res = TEST_REASONING_RESULT_BAD;
	}

	if(pResult->k < 0.023f || pResult->k > 0.03f)
	{
		pReasoningResult->emf = TEST_REASONING_RESULT_BAD;
	}

	if(pResult->cur < 0.06f || pResult->cur > 0.1f)
	{
		pReasoningResult->cur = TEST_REASONING_RESULT_WARNING;
	}

	if(pResult->cur < 0.03f || pResult->cur > 0.2f)
	{
		pReasoningResult->cur = TEST_REASONING_RESULT_BAD;
	}

	if(pResult->speed < 210 || pResult->speed > 240)
	{
		pReasoningResult->vel = TEST_REASONING_RESULT_WARNING;
	}

	if(pResult->speed < 150 || pResult->speed > 300)
	{
		pReasoningResult->vel = TEST_REASONING_RESULT_BAD;
	}
}

void calculateReasoningOnTestResultsDribbler(const TestMotorDynamicsResult* pResult, TestDynamicsReasoningResult* pReasoningResult) {
	if(pResult->R < 0.8f || pResult->R > 1.1f)
	{
		pReasoningResult->res = TEST_REASONING_RESULT_WARNING;
	}

	if(pResult->R < 0.6f || pResult->R > 1.3f)
	{
		pReasoningResult->res = TEST_REASONING_RESULT_BAD;
	}

	if(pResult->k < 0.008f || pResult->k > 0.012f)
	{
		pReasoningResult->emf = TEST_REASONING_RESULT_BAD;
	}

	if(pResult->cur < 0.20f || pResult->cur > 0.30f)
	{
		pReasoningResult->cur = TEST_REASONING_RESULT_WARNING;
	}

	if(pResult->cur < 0.1f || pResult->cur > 0.5f)
	{
		pReasoningResult->cur = TEST_REASONING_RESULT_BAD;
	}

	if(pResult->speed < 620 || pResult->speed > 660)
	{
		pReasoningResult->vel = TEST_REASONING_RESULT_WARNING;
	}

	if(pResult->speed < 580 || pResult->speed > 700)
	{
		pReasoningResult->vel = TEST_REASONING_RESULT_BAD;
	}
}

void calculateReasoningOnTestResults(const TestMotorDynamicsResult* pResult, TestDynamicsReasoningResult* pReasoningResult) {
	pReasoningResult->cur = TEST_REASONING_RESULT_OK;
	pReasoningResult->res = TEST_REASONING_RESULT_OK;
	pReasoningResult->emf = TEST_REASONING_RESULT_OK;
	pReasoningResult->vel = TEST_REASONING_RESULT_OK;
	if (pResult->motorId < 4)
		calculateReasoningOnTestResultsDrive(pResult, pReasoningResult);
	else
		calculateReasoningOnTestResultsDribbler(pResult, pReasoningResult);

}

inline void printReasoningResultConstant(uint8_t result) {
	if (result == TEST_REASONING_RESULT_OK) {
		ConsolePrint("OK");
	} else if (result == TEST_REASONING_RESULT_WARNING) {
		ConsolePrint("WARNING");
	} else if (result == TEST_REASONING_RESULT_BAD) {
		ConsolePrint("BAD");
	}

}

inline void printTestMotorDynamicsResult(TestMotorDynamicsResult* pResult) {
	ConsolePrint("Finished dynamics test on motor %hu", pResult->motorId);
	if (pResult->motorId < 4) {
		ConsolePrint(" (drive motor)\r\n");
	} else {
		ConsolePrint(" (dribbler motor)\r\n");
	}
	TestDynamicsReasoningResult reasoningResult;
	calculateReasoningOnTestResults(pResult, &reasoningResult);
	ConsolePrint("\tres: R=%f (", pResult->R);
	printReasoningResultConstant(reasoningResult.res);
	ConsolePrint(")\n\r");
	ConsolePrint("\temf: k=%f (", pResult->k);
	printReasoningResultConstant(reasoningResult.emf);
	ConsolePrint(")\n\r");
	ConsolePrint("\tcur: cur=%f (", pResult->cur);
	printReasoningResultConstant(reasoningResult.cur);
	ConsolePrint(")\n\r");
	ConsolePrint("\tvel: speed=%f (", pResult->speed);
	printReasoningResultConstant(reasoningResult.vel);
	ConsolePrint(")\n\r");
}

