/*
 * test.c
 *
 *  Created on: 13.12.2015
 *      Author: AndreR <andre@ryll.cc>
 */

#include "test.h"
#include "motors.h"
#include "util/intercom.h"
#include "util/intercom_reliable.h"
#include "util/ema_filter.h"
#include "util/arm_mat_util_f32.h"

Test test;

static const uint8_t hallSortLUT[8] = { 0, 5, 1, 0, 3, 4, 2, 0 };

static void testCallback(const IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData, void* pUserData)
{
	(void)pUserData;

	if(pHeader->command == IC_TEST_START && dataLength >= sizeof(TestStart))
	{
		TestStart* pStart = (TestStart*)pData;
#ifdef MOT_RIGHT
		if(pStart->motorId == 0 || pStart->motorId == 3)
#else
		if(pStart->motorId == 1 || pStart->motorId == 2)
#endif
		{
			msg_t event = pStart->motorId;
			event <<= 8;
			event |= pStart->testId;

			chMBPost(&test.eventQueue, event, TIME_IMMEDIATE);
		}
	}
}

void TestInit()
{
	chMBObjectInit(&test.eventQueue, test.eventQueueData, TEST_EVENT_QUEUE_SIZE);

	IntercomClient client;
	client.pUserData = 0;
	client.rxCallback = &testCallback;
	client.section = INTERCOM_SECTION_TEST;

	IntercomAddClient(client);
}

void TestTask(void* params)
{
	(void)params;

	msg_t event;
	IntercomHeader iHeader;
	iHeader.address = INTERCOM_ADDRESS_MEDIA;
	iHeader.section = INTERCOM_SECTION_TEST;

	chRegSetThreadName("Test");

	while(1)
	{
		if(chMBFetch(&test.eventQueue, &event, TIME_INFINITE) != MSG_OK)
			continue;

		uint8_t testId = event & 0xFF;
		uint8_t motorId = event >> 8;

		Drive* pDrive = &drive[0];
		if(motorId > 1)
			pDrive = &drive[1];

		switch(testId)
		{
			case IC_TEST_MOT_HALL:
			{
				TestMotorHallResult result;
				TestMotorHall(pDrive, &result);

				iHeader.command = IC_TEST_MOT_HALL;
				IntercomReliableSend(&iHeader, sizeof(TestMotorHallResult), &result);
			}
			break;
			case IC_TEST_MOT_POWER:
			{
				TestMotorPowerUnitResult result;
				TestMotorPowerUnit(pDrive, 1.0f, &result);

				iHeader.command = IC_TEST_MOT_POWER;
				IntercomReliableSend(&iHeader, sizeof(TestMotorHallResult), &result);
			}
			break;
			case IC_TEST_MOT_DYN:
			{
				TestMotorDynamicsResult result;
				TestMotorDynamics(pDrive, &result);

				iHeader.command = IC_TEST_MOT_DYN;
				IntercomReliableSend(&iHeader, sizeof(TestMotorHallResult), &result);
			}
			break;
			case IC_TEST_MOT_ENC:
			{
				TestMotorEncoderResult result;
				TestMotorEncoder(pDrive, &result);

				iHeader.command = IC_TEST_MOT_ENC;
				IntercomReliableSend(&iHeader, sizeof(TestMotorHallResult), &result);
			}
			break;
		}
	}
}

void TestMotorHall(Drive* pDrive, TestMotorHallResult* pResult)
{
	TestProgress prog;
	prog.testId = IC_TEST_MOT_HALL;
	prog.motorId = pDrive->id;
	prog.percentComplete = 0;

	IntercomHeader iHeader;
	iHeader.address = INTERCOM_ADDRESS_MEDIA;
	iHeader.section = INTERCOM_SECTION_TEST;
	iHeader.command = IC_TEST_PROGRESS;

	IntercomSend(&iHeader, sizeof(TestProgress), &prog);

	memset(pResult, 0, sizeof(TestMotorHallResult));
	pResult->motorId = pDrive->id;

	motors.ignoreMainCmds = 1;
	chThdSleepMilliseconds(5);

	pDrive->slopeLimited = 0;
	DriveSetCtrlMode(pDrive, DRIVE_CTRL_MODE_NONE);

	uint16_t hall = pDrive->halPos;
	float v = 1.0f;

	for(float a = 0; a < 360; a += 1.0f)
	{
		DriveSetVoltageAngle(pDrive, v, a);

		prog.percentComplete = a/361.0f*50.0f;
		IntercomSend(&iHeader, sizeof(TestProgress), &prog);

		chThdSleepMilliseconds(10);

		if(hall != pDrive->halPos)
		{
			hall = pDrive->halPos;
			pResult->pos[hallSortLUT[hall]].fwd = (int16_t)a;
		}
	}

	prog.percentComplete = 50;
	IntercomSend(&iHeader, sizeof(TestProgress), &prog);

	for(float a = 360; a > 0; a -= 1.0f)
	{
		DriveSetVoltageAngle(pDrive, v, a);

		prog.percentComplete = 50.0f + (360-a)/361.0f*50.0f;
		IntercomSend(&iHeader, sizeof(TestProgress), &prog);

		chThdSleepMilliseconds(10);

		if(hall != pDrive->halPos)
		{
			hall = pDrive->halPos;
			pResult->pos[hallSortLUT[hall]].rev = (int16_t)a;
		}
	}

	DriveSetVoltage(pDrive, 0);

	for(uint8_t i = 0; i < 6; i++)
		pResult->pos[i].ref = i*60;

	prog.percentComplete = 100;
	IntercomSend(&iHeader, sizeof(TestProgress), &prog);

	chThdSleepMilliseconds(50);

	pDrive->slopeLimited = 1;

	motors.ignoreMainCmds = 0;
}

void TestMotorPowerUnit(Drive* pDrive, float voltage, TestMotorPowerUnitResult* pResult)
{
	EMAFilter mainCurrent;
	EMAFilter motCurrent;

	EMAFilterInit(&mainCurrent, 0.95f);
	EMAFilterInit(&motCurrent, 0.95f);

	TestProgress prog;
	prog.testId = IC_TEST_MOT_POWER;
	prog.motorId = pDrive->id;
	prog.percentComplete = 0;

	IntercomHeader iHeader;
	iHeader.address = INTERCOM_ADDRESS_MEDIA;
	iHeader.section = INTERCOM_SECTION_TEST;
	iHeader.command = IC_TEST_PROGRESS;

	IntercomSend(&iHeader, sizeof(TestProgress), &prog);

	pResult->motorId = pDrive->id;

	motors.ignoreMainCmds = 1;
	chThdSleepMilliseconds(5);

	pDrive->slopeLimited = 0;

	DriveSetCtrlMode(pDrive, DRIVE_CTRL_MODE_NONE);
	DriveSetMode(pDrive, DRIVE_MODE_DISABLE);

	for(uint16_t j = 0; j < 1000; j++)
	{
		EMAFilterUpdate(&mainCurrent, pDrive->cur);
		chThdSleepMilliseconds(1);
	}

	float mainOffset = mainCurrent.value;

	prog.percentComplete = 25;
	IntercomSend(&iHeader, sizeof(TestProgress), &prog);

//	ConsolePrint("Main current offset: %d\r\n", (int32_t)(mainOffset*1000.0f));

	DriveSetVoltage(pDrive, voltage);

	for(uint16_t ph = 0; ph < 3; ph++)
	{
		DriveSetMode(pDrive, DRIVE_MODE_TEST1+ph);
		chThdSleepMilliseconds(100);

		for(uint16_t j = 0; j < 1000; j++)
		{
			EMAFilterUpdate(&mainCurrent, pDrive->cur-mainOffset);
			EMAFilterUpdate(&motCurrent, pDrive->rmsCurrent);
			chThdSleepMilliseconds(1);
		}

		float mainCur = pDrive->bat*mainCurrent.value/voltage;

		pResult->mainCur[ph] = mainCur;
		pResult->motCur[ph] = motCurrent.value;

		prog.percentComplete = 50+ph*25;
		IntercomSend(&iHeader, sizeof(TestProgress), &prog);
	}

	DriveSetVoltage(pDrive, 0.0f);
	DriveSetMode(pDrive, DRIVE_MODE_DISABLE);

	float iA = pResult->mainCur[0];
	float iB = pResult->mainCur[1];
	float iC = pResult->mainCur[2];
	float R1 = 0.5f*(1.0f/iC-1.0f/iB+1.0f/iA);
	float R2 = 1.0f/iA-R1;
	float R3 = 1.0f/iC-R1;

	pResult->R[0] = R1;
	pResult->R[1] = R2;
	pResult->R[2] = R3;

	chThdSleepMilliseconds(50);

	pDrive->slopeLimited = 1;

	motors.ignoreMainCmds = 0;
}

void TestMotorEncoder(Drive* pDrive, TestMotorEncoderResult* pResult)
{
	TestProgress prog;
	prog.testId = IC_TEST_MOT_ENC;
	prog.motorId = pDrive->id;
	prog.percentComplete = 0;

	IntercomHeader iHeader;
	iHeader.address = INTERCOM_ADDRESS_MEDIA;
	iHeader.section = INTERCOM_SECTION_TEST;
	iHeader.command = IC_TEST_PROGRESS;

	IntercomSend(&iHeader, sizeof(TestProgress), &prog);

	pResult->motorId = pDrive->id;

	motors.ignoreMainCmds = 1;
	chThdSleepMilliseconds(5);

	pDrive->slopeLimited = 0;

	DriveSetCtrlMode(pDrive, DRIVE_CTRL_MODE_NONE);
	DriveSetMode(pDrive, DRIVE_MODE_SVPWM);

	float power = 1.0f;

	DriveSetVoltageAngle(pDrive, power, 0);

	chThdSleepMilliseconds(50);

	uint16_t start = pDrive->pTIMEncoder->CNT;

	for(float a = 0; a <= 5*8*360; a += 5)
	{
		DriveSetVoltageAngle(pDrive, power, a);

		chThdSleepMilliseconds(2);

		if(((uint16_t)a)%60 == 0)
		{
			prog.percentComplete = a/144.0f;
			IntercomSend(&iHeader, sizeof(TestProgress), &prog);
		}
	}

	uint16_t end = pDrive->pTIMEncoder->CNT;

	pResult->measured = end-start;
	pResult->expected = 5*2048;

	DriveSetVoltageAngle(pDrive, 0, 0);

	chThdSleepMilliseconds(50);

	pDrive->slopeLimited = 1;

	motors.ignoreMainCmds = 0;
}

#define EST_SIZE 16
void TestMotorDynamics(Drive* pDrive, TestMotorDynamicsResult* pResult)
{
	EMAFilter mainCurrent;
	EMAFilter current;
	EMAFilter speed;

	arm_matrix_instance_f32 A = {EST_SIZE, 2, (float[EST_SIZE*2]){}};
	arm_matrix_instance_f32 Ai = {2, EST_SIZE, (float[EST_SIZE*2]){}};
	arm_matrix_instance_f32 b = {EST_SIZE, 1, (float[EST_SIZE*1]){}};
	arm_matrix_instance_f32 x = {2, 1, (float[2]){}};

	TestProgress prog;
	prog.testId = IC_TEST_MOT_DYN;
	prog.motorId = pDrive->id;
	prog.percentComplete = 0;

	IntercomHeader iHeader;
	iHeader.address = INTERCOM_ADDRESS_MEDIA;
	iHeader.section = INTERCOM_SECTION_TEST;
	iHeader.command = IC_TEST_PROGRESS;

	IntercomSend(&iHeader, sizeof(TestProgress), &prog);

	pResult->motorId = pDrive->id;

	motors.ignoreMainCmds = 1;
	chThdSleepMilliseconds(5);

	pDrive->slopeLimited = 0;

	arm_mat_zero_f32(&A);
	arm_mat_zero_f32(&Ai);
	arm_mat_zero_f32(&b);
	arm_mat_zero_f32(&x);

	EMAFilterInit(&mainCurrent, 0.5f);
	EMAFilterInit(&current, 0.95f);
	EMAFilterInit(&speed, 0.95f);

	DriveSetCtrlMode(pDrive, DRIVE_CTRL_MODE_NONE);
	DriveSetMode(pDrive, DRIVE_MODE_BLOCK);

	for(uint16_t j = 0; j < 1000; j++)
	{
		EMAFilterUpdate(&mainCurrent, pDrive->cur);
		chThdSleepMilliseconds(1);
	}

	float mainOffset = mainCurrent.value;

	prog.percentComplete = 1/18.0f*100.0f;
	IntercomSend(&iHeader, sizeof(TestProgress), &prog);

//	ConsolePrint("Main current offset: %d\r\n", (int32_t)(mainOffset*1000.0f));

	for(int16_t i = 0; i <= EST_SIZE; i++)
	{
		float volt = i-EST_SIZE/2;

		if(i == EST_SIZE/2)
			continue;

		DriveSetVoltage(pDrive, volt);
		chThdSleepMilliseconds(200);

		for(uint16_t j = 0; j < 800; j++)
		{
			EMAFilterUpdate(&mainCurrent, pDrive->cur-mainOffset);
			EMAFilterUpdate(&current, pDrive->rmsCurrent);
			EMAFilterUpdate(&speed, pDrive->curVelocity*2*M_PI);

			chThdSleepMilliseconds(1);
		}

//		float mainCur = pDrive->bat*mainCurrent.value/volt;

		if(volt < 0)
			MAT_ELEMENT(A, i-1, 0) = -current.value;
		else
			MAT_ELEMENT(A, i-1, 0) = current.value;
		MAT_ELEMENT(A, i-1, 1) = speed.value;
		MAT_ELEMENT(b, i-1, 0) = volt;

		prog.percentComplete = ((float)(i+2))/18.0f*100.0f;
		IntercomSend(&iHeader, sizeof(TestProgress), &prog);

//		ConsolePrint("U: %dmV, I: %dmA/%dmA, v: %dmrad/s\r\n",
//				(int32_t)(volt*1000.0f), (int32_t)(current.value*1000.0f), (int32_t)(mainCur*1000.0f), (int32_t)(speed.value*1000.0f));
	}

	DriveSetVoltage(pDrive, 0);

	arm_mat_pinv(&A, &Ai);
	arm_mat_mult_f32(&Ai, &b, &x);

	float R = MAT_ELEMENT(x, 0, 0);
	float k = MAT_ELEMENT(x, 1, 0);

	pResult->R = R;
	pResult->k = k;
	pResult->cur = current.value;
	pResult->speed = speed.value;

	prog.percentComplete = 100;
	IntercomSend(&iHeader, sizeof(TestProgress), &prog);

	pDrive->slopeLimited = 1;

	motors.ignoreMainCmds = 0;

	chThdSleepMilliseconds(1000);
}
