/*
 * log_msgs.h
 *
 *  Created on: 21.10.2014
 *      Author: AndreR
 */

#ifndef LOG_MSGS_H_
#define LOG_MSGS_H_

#include <stdint.h>
#include "util/pid.h"

#ifndef PACKED
#define PACKED __attribute__((packed))
#endif

// fundamental header for all entries in the log file
typedef struct PACKED _LogChunkHeader
{
	uint16_t type;
	uint16_t length;
} LogChunkHeader;

// the header for all log entries, contains the LogChunkHeader
typedef struct PACKED _LogEntryHeader
{
	uint16_t type;
	uint16_t length;
	uint32_t timestamp;
} LogEntryHeader;

// --- LOG DEFINITIONS ---
typedef struct PACKED _LogRawI16XYZ
{
	LogEntryHeader header;
	int16_t xyz[3];
} LogRawI16XYZ;

#define MOT_FR 0
#define MOT_FL 1
#define MOT_RL 2
#define MOT_RR 3

typedef struct PACKED _RobotSensors
{
	LogEntryHeader header;

	struct PACKED
	{
		uint32_t time;
		float vel[4];		// angular motor velocity [rad/s]
	} enc;

	struct PACKED
	{
		uint32_t updated;
		uint32_t time;
		float rotVel[3];	// around X, Y, Z [rad/s]
	} gyr;

	struct PACKED
	{
		uint32_t updated;
		uint32_t time;
		float linAcc[3];	// X, Y, Z [m^2/s]
	} acc;

	struct PACKED
	{
		uint32_t updated;
		uint32_t time;		// local arrival time (systime)
		uint32_t delay;		// t [us]
		float pos[3];		// X, Y, Z [m]
		uint32_t camId;
		uint32_t noVision;
	} vision;

	struct PACKED
	{
		float voltage;		// battery voltage [V]
		float current;		// main current consumption [A]
		uint32_t batEmpty;
	} power;

	struct PACKED
	{
		uint32_t time;
		float vol[4];		// [V]
	} motorVol;

	struct PACKED
	{
		float level;		// [V]
		float chgCurrent;	// [A]
		uint32_t kickCounter;
		uint32_t flags;
	} kicker;

	struct PACKED
	{
		float speed;		// [RPM]
		float temp;			// [C]
		float voltage;
		uint32_t overheated;
	} dribbler;

	struct PACKED
	{
		float level[2];		// (on|off) [V]
		uint32_t interrupted;
	} ir;
} RobotSensors;

typedef struct PACKED _RobotCtrlState
{
	LogEntryHeader header;

	float pos[3];	// [m]
	float vel[3];	// [m/s]
	float acc[3];

	uint32_t posUpdated;
	uint32_t velUpdated;
	uint32_t accUpdated;
} RobotCtrlState;

typedef struct PACKED _RobotCtrlOutput
{
	LogEntryHeader header;

	float motorVel[4];	// angular motor velocity [rad/s]
	float motorVoltage[4];	// Motor voltage [V], -CCW, +CW
	float motorAcc[4];	// angular acceleration [rad/s^2]. Used together with velocity
	uint32_t ctrlMode;
} RobotCtrlOutput;

// output from trajectory generator
typedef struct PACKED _RobotCtrlReference
{
	LogEntryHeader header;

	float trajPos[3];
	float trajVelLocal[3];
	float trajAccLocal[3];
	float trajJerkLocal[3];
} RobotCtrlReference;

typedef struct PACKED _RobotDriveLimits
{
	float velMaxXY;
	float accMaxXY;
	float jerkMaxXY;

	float velMaxW;
	float accMaxW;
	float jerkMaxW;
} RobotDriveLimits;

typedef struct PACKED _LogCtrlTiggaDebug
{
	LogEntryHeader header;

	uint32_t fusionTime;
	uint32_t ctrlTime;
} LogCtrlTiggaDebug;

typedef struct PACKED _LogOpticalRaw
{
	LogEntryHeader header;

	int16_t lDeltaX;
	int16_t lDeltaY;
	uint8_t lSqual;
	int16_t rDeltaX;
	int16_t rDeltaY;
	uint8_t rSqual;
} LogOpticalRaw;

typedef struct PACKED _DriveInput
{
	float pos[3];		// [m]
	float localVel[3];	// [m/s]
	float wheelVel[4];	// [rad/s] @ wheel
	uint32_t modeXY;
	uint32_t modeW;

	RobotDriveLimits limits;
} DriveInput;

typedef struct PACKED _DribblerInput
{
	float speed;		// [RPM]
	float voltage;		// [V]
	uint32_t mode;
} DribblerInput;

typedef struct PACKED _KickerInput
{
	uint32_t mode;
	uint32_t device;
	float speed;		// [m/s]
} KickerInput;

typedef struct PACKED _SkillOutput
{
	LogEntryHeader header;

	DriveInput drive;
	DribblerInput dribbler;
	KickerInput kicker;
} SkillOutput;

// LOGFILE descriptions
enum ElementType
{
	UINT8 = 0,
	INT8,
	UINT16,
	INT16,
	UINT32,
	INT32,
	FLOAT,
	DOUBLE,
};

// describes a single structure element
typedef struct PACKED _ElementDesc
{
	uint8_t type;
	const char* pName;	// 60 characters max
	const char* pUnit;
	const char* pDesc;
} ElementDesc;

// describes a message
typedef struct PACKED _LogMessageDesc
{
	uint16_t msgId;
	const char* pName;
	uint16_t numElements;
	ElementDesc* elements;
} LogMessageDesc;

extern const LogMessageDesc logMessageDescriptions[9];

#endif /* LOG_MSGS_H_ */
