/*
 * ctrl_tigga.c
 *
 *  Created on: 26.02.2015
 *      Author: AndreR
 */

#include <main/ctrl.h>
#include <main/ctrl_tigga.h>
#include "struct_ids.h"
#include "util/arm_mat_util_f32.h"
#include "util/angle_math.h"
#include "util/sys_time.h"
#include "util/log.h"
#include "util/console.h"
#include <float.h>

// DEFINITIONS

// FORWARD DECLARATIONS
static void init(const float* pDefaultPos);
static void controller(const RobotCtrlState* pState, const DriveInput* pDrive, RobotCtrlOutput* pOutput, RobotCtrlReference* pReference);
static void sensorFusion(const RobotSensors* pSensors, const RobotCtrlOutput* pOutput, RobotCtrlState* pState);

static void ekfStateFunc(arm_matrix_instance_f32* pX, const arm_matrix_instance_f32* pU);
static void ekfStateJacobianFunc(const arm_matrix_instance_f32* pX, const arm_matrix_instance_f32* pU, arm_matrix_instance_f32* pF);
static void ekfMeasFunc(const arm_matrix_instance_f32* pX, arm_matrix_instance_f32* pY);
static void ekfMeasJacobianFunc(const arm_matrix_instance_f32* pX, arm_matrix_instance_f32* pH);

static inline const float* visionProc(uint8_t updated, int32_t tDelayUs, const float* pVisPos);

// GLOBAL DATA
CtrlTigga ctrlTigga = {
	.configFilter = {
		.posNoiseXY = 0.001f,
		.posNoiseW = 0.001f,
		.velNoiseXY = 0.03f,
		.accBiasNoiseXY = 0.00003f,
		.gyrBiasNoiseW = 0.00001f,
		.visNoiseXY = 0.003f,
		.visNoiseW = 0.05f,
		.outlierMaxVelXY = 5.0f,
		.outlierMaxVelW = 50.0f,
		.trackingCoeff = 2.0f,
		.visCaptureDelay = 25,
		.visProcessingDelayMax = 20,
		.visionTimeoutMs = 1000,
		.encPosRate = 50,
	},
	.configMisc = {
		.fieldSizeX = 10.4f,
		.fieldSizeY = 7.4,
		.ASR = 1,
		.useFeedforward = 1,
	},
	.configPid = {
		.posXY = {8.0f, 0.0f, 0.1f},
		.posW = {5.0f, 0, 0.1f},
		.velXY = {1, 0, 0},
		.velW = {0, 20, 0},
	},
	.configBotModel = {
		.X = {0.21156270f,  0.06147342f, -0.02937017f},
		.Y = {0.05196181f, -0.04677604f, -0.01003764f},
	},
};

CtrlInstance ctrlTiggaInstance = {
	.pInit = &init,
	.pExit = 0,
	.pStateEstimationFunc = &sensorFusion,
	.pControllerFunc = &controller,
	.pLog = &ctrlTigga.log,
};

// LOCAL DATA - CONFIGS
static const ConfigFileDesc configFileDescFilter =
	{ SID_CFG_CTRL_TIGGA_FILTER, 33, "ctrl_tigga/filter", 14, (ElementDesc[]) {
		{ FLOAT, "noise_pos_xy", "m", "noise/pos/xy" },
		{ FLOAT, "noise_pos_w", "rad", "noise/pos/w" },
		{ FLOAT, "noise_vel_xy", "m/s", "noise/vel/xy" },
		{ FLOAT, "noise_acc_bias", "m/s^2", "noise/bias/acc" },
		{ FLOAT, "noise_gyr_bias", "m/s^2", "noise/bias/gyr" },
		{ FLOAT, "noise_vis_xy", "m", "noise/vis/xy" },
		{ FLOAT, "noise_vis_w", "rad", "noise/vis/w" },
		{ FLOAT, "outlier_max_vel_xy", "m/s", "outlier/vel/xy" },
		{ FLOAT, "outlier_max_vel_w", "rad/s", "outlier/vel/w" },
		{ FLOAT, "tracking_coeff", "", "tracking/coeff" },
		{ UINT8, "vis_capture_delay", "ms", "vision/capture/delay" },
		{ UINT8, "vis_prc_delay_max", "ms", "vision/proc/delay (max)" },
		{ UINT16, "vision_timeout_ms", "ms", "vision/timeout" },
		{ UINT16, "enc_pos_rate", "Hz", "encoder/position/rate" },
	} };

static const ConfigFileDesc configFileDescMisc =
	{ SID_CFG_CTRL_TIGGA_MISC, 14, "ctrl_tigga/misc", 4, (ElementDesc[]) {
		{ FLOAT, "field_size_x", "m", "field/size/x" },
		{ FLOAT, "field_size_y", "m", "field/size/y" },
		{ UINT8, "asr", "", "ASR" },
		{ UINT8, "feedforward", "", "Use Feedforward" },
	} };

static const ConfigFileDesc configFileDescPid =
	{ SID_CFG_CTRL_TIGGA_PID, 5, "ctrl_tigga/pid", 12, (ElementDesc[]) {
		{ FLOAT, "pos_xy_p", "", "pos/xy/p" },
		{ FLOAT, "pos_xy_i", "", "pos/xy/i" },
		{ FLOAT, "pos_xy_d", "", "pos/xy/d" },
		{ FLOAT, "pos_w_p", "", "pos/w/p" },
		{ FLOAT, "pos_w_i", "", "pos/w/i" },
		{ FLOAT, "pos_w_d", "", "pos/w/d" },
		{ FLOAT, "vel_xy_p", "", "vel/xy/p" },
		{ FLOAT, "vel_xy_i", "", "vel/xy/i" },
		{ FLOAT, "vel_xy_d", "", "vel/xy/d" },
		{ FLOAT, "vel_w_p", "", "vel/w/p" },
		{ FLOAT, "vel_w_i", "", "vel/w/i" },
		{ FLOAT, "vel_w_d", "", "vel/w/d" },
	} };

static const ConfigFileDesc configFileDescBotModel =
	{ SID_CFG_CTRL_TIGGA_BOT_MODEL, 1, "ctrl_tigga/bot_model", 6, (ElementDesc[]) {
		{ FLOAT, "x1", "", "cos(1*x)" },
		{ FLOAT, "x3", "", "cos(3*x)" },
		{ FLOAT, "x5", "", "cos(5*x)" },
		{ FLOAT, "y1", "", "sin(1*y)" },
		{ FLOAT, "y3", "", "sin(3*y)" },
		{ FLOAT, "y5", "", "sin(5*y)" },
	} };

static void configUpdateCallback(uint16_t cfgId)
{
	switch(cfgId)
	{
		case SID_CFG_CTRL_TIGGA_FILTER:
		{
			uint16_t bufSize = (uint16_t)ctrlTigga.configFilter.visCaptureDelay + (uint16_t)ctrlTigga.configFilter.visProcessingDelayMax;
			if(bufSize > CTRL_TIGGA_MAX_BUFFER_SIZE)
				bufSize = CTRL_TIGGA_MAX_BUFFER_SIZE;

			ctrlTigga.vision.size = bufSize;
			ctrlTigga.gyrAcc.size = bufSize;
			ctrlTigga.enc.size = bufSize;
			ctrlTigga.ins.size = bufSize;

			ctrlTigga.vision.readPtr = 0;
			ctrlTigga.gyrAcc.writePtr = 0;
			ctrlTigga.enc.writePtr = 0;
			ctrlTigga.ins.writePtr = 0;

			MAT_ELEMENT(ctrlTigga.ekf.Ex, 0, 0) = ctrlTigga.configFilter.posNoiseXY*ctrlTigga.configFilter.posNoiseXY;
			MAT_ELEMENT(ctrlTigga.ekf.Ex, 1, 1) = ctrlTigga.configFilter.posNoiseXY*ctrlTigga.configFilter.posNoiseXY;
			MAT_ELEMENT(ctrlTigga.ekf.Ex, 2, 2) = ctrlTigga.configFilter.posNoiseW*ctrlTigga.configFilter.posNoiseW;
			MAT_ELEMENT(ctrlTigga.ekf.Ex, 3, 3) = ctrlTigga.configFilter.velNoiseXY*ctrlTigga.configFilter.velNoiseXY;
			MAT_ELEMENT(ctrlTigga.ekf.Ex, 4, 4) = ctrlTigga.configFilter.velNoiseXY*ctrlTigga.configFilter.velNoiseXY;

			MAT_ELEMENT(ctrlTigga.ekf.Ez, 0, 0) = ctrlTigga.configFilter.visNoiseXY*ctrlTigga.configFilter.visNoiseXY;
			MAT_ELEMENT(ctrlTigga.ekf.Ez, 1, 1) = ctrlTigga.configFilter.visNoiseXY*ctrlTigga.configFilter.visNoiseXY;
			MAT_ELEMENT(ctrlTigga.ekf.Ez, 2, 2) = ctrlTigga.configFilter.visNoiseW*ctrlTigga.configFilter.visNoiseW;
		}
		break;
		case SID_CFG_CTRL_TIGGA_PID:
		{
			PIDSetParams(&ctrlTigga.posPIDXY, ctrlTigga.configPid.posXY.kp, ctrlTigga.configPid.posXY.ki, ctrlTigga.configPid.posXY.kd);
			PIDSetParams(&ctrlTigga.posPIDW, ctrlTigga.configPid.posW.kp, ctrlTigga.configPid.posW.ki, ctrlTigga.configPid.posW.kd);

			PIDSetParams(&ctrlTigga.velPIDXY, ctrlTigga.configPid.velXY.kp, ctrlTigga.configPid.velXY.ki, ctrlTigga.configPid.velXY.kd);
			PIDSetParams(&ctrlTigga.velPIDW, ctrlTigga.configPid.velW.kp, ctrlTigga.configPid.velW.ki, ctrlTigga.configPid.velW.kd);
		}
		break;
		case SID_CFG_CTRL_TIGGA_MISC:
		{
		}
		break;
	}
}

static void init(const float* pDefaultPos)
{
	const float dt = CTRL_DELTA_T;

	ctrlTigga.log.header.type = SID_CTRL_TIGGA_DEBUG;
	ctrlTigga.log.header.length = sizeof(LogCtrlTiggaDebug);

	ctrlTigga.visLost = 1;
	ctrlTigga.filterUnstable = 1;
	ctrlTigga.lastVisionTime = 0;
	ctrlTigga.noVisionCounter = 0;
	ctrlTigga.motorOffCounter = 0;

	ctrlTigga.numLateVisionMeasurements = 0;
	ctrlTigga.maxVisionDelay = 0;

	ctrlTigga.lastInitTime = chVTGetSystemTimeX();

	memset(ctrlTigga.mainState, 0, sizeof(float)*8);
	memset(ctrlTigga.mechState, 0, sizeof(float)*8);
	memcpy(ctrlTigga.mainState, pDefaultPos, sizeof(float)*3);

	memset(&ctrlTigga.vision, 0, sizeof(ctrlTigga.vision));
	memset(&ctrlTigga.gyrAcc, 0, sizeof(ctrlTigga.gyrAcc));
	memset(&ctrlTigga.enc, 0, sizeof(ctrlTigga.enc));
	memset(&ctrlTigga.ins, 0, sizeof(ctrlTigga.ins));

	memset(ctrlTigga.bias, 0, sizeof(ctrlTigga.bias));

	memset(&ctrlTigga.log, 0, sizeof(ctrlTigga.log));

	ctrlTigga.lastModeXY = DRIVE_MODE_OFF;
	ctrlTigga.lastModeW = DRIVE_MODE_OFF;

	// FILTER
	EKFInit(&ctrlTigga.ekf, 5, 3, 3, ctrlTigga.ekfData);
	arm_mat_scale_f32(&ctrlTigga.ekf.Sigma, 0.001f, &ctrlTigga.ekf.Sigma);
	ctrlTigga.ekf.pState = &ekfStateFunc;
	ctrlTigga.ekf.pStateJacobian = &ekfStateJacobianFunc;
	ctrlTigga.ekf.pMeas = &ekfMeasFunc;
	ctrlTigga.ekf.pMeasJacobian = &ekfMeasJacobianFunc;

	arm_mat_identity_f32(&ctrlTigga.ekf.Ex);
	arm_mat_identity_f32(&ctrlTigga.ekf.Ez);

	memcpy(ctrlTigga.ekf.x.pData, pDefaultPos, sizeof(float)*3);

	// CONTROL
	EMAFilterInit(&ctrlTigga.emaCorrection[0], 0.8f);
	EMAFilterInit(&ctrlTigga.emaCorrection[1], 0.8f);

	EMAFilterInit(&ctrlTigga.emaBias[0], 0.98f);
	EMAFilterInit(&ctrlTigga.emaBias[1], 0.98f);
	EMAFilterInit(&ctrlTigga.emaBias[2], 0.98f);

	PIDParams pp;
	pp.kp = 1.0f;
	pp.kd = 0;
	pp.ki = 0;
	pp.outputMin = -100;
	pp.outputMax = 100;
	PIDInit(&ctrlTigga.velPIDW, &pp, CTRL_DELTA_T_IN_US*1e-6f, 1);

	pp.outputMin = -5.0f;
	pp.outputMax = 5.0f;
	PIDInit(&ctrlTigga.velPIDXY, &pp, CTRL_DELTA_T_IN_US*1e-6f, 1);

	pp.outputMin = -5.0f;
	pp.outputMax = 5.0f;
	PIDInit(&ctrlTigga.posPIDXY, &pp, dt, 1);

	pp.outputMin = -50.0f;
	pp.outputMax = 50.0f;
	PIDInit(&ctrlTigga.posPIDW, &pp, dt, 1);

	TrajGeneratorInit(&ctrlTigga.trajGen);

	ConfigNotifyUpdate(ctrlTigga.pConfigFileFilter);
	ConfigNotifyUpdate(ctrlTigga.pConfigFilePid);
	ConfigNotifyUpdate(ctrlTigga.pConfigFileMisc);
}

static inline void updatePID2D(PID* pPID, const float* pSet, const float* pMeas, float* pOut, uint8_t firstRun, float maxErr)
{
	// XY position control
	float errorPosX = pMeas[0] - pSet[0];
	float errorPosY = pMeas[1] - pSet[1];
	float errorXY = sqrtf(errorPosX*errorPosX+errorPosY*errorPosY);
	if(errorXY > maxErr)
		errorXY = maxErr;

	// process PIDs
	pPID->setpoint = 0;
	if(firstRun)
		pPID->lastMeasurement = errorXY; // prevent derivative kick
	PIDUpdate(pPID, errorXY);

	if(errorXY > 0.0001f)
	{
		pOut[0] = errorPosX/errorXY * pPID->output;
		pOut[1] = errorPosY/errorXY * pPID->output;
	}
	else
	{
		pOut[0] = 0.0f;
		pOut[1] = 0.0f;
	}
}

static inline void xyCorrection(const float* pIn, float* pCor)
{
	float* a = ctrlTigga.configBotModel.X;
	float* b = ctrlTigga.configBotModel.Y;

	float x = atan2f(pIn[1], pIn[0]);

	pCor[0] = a[0]*cosf(x) + a[1]*cosf(3*x) + a[2]*cosf(5*x);
	pCor[1] = b[0]*sinf(x) + b[1]*sinf(3*x) + b[2]*sinf(5*x);
}

static inline void velCorrection(const float* pVel, const float* pAcc, float* pCorr)
{
	const float cruiseWindow = 0.1f; // [rad]
	const float piHalf = M_PI*0.5f;

	float velCor[2];
	xyCorrection(pVel, velCor);

	float accCor[2];
	xyCorrection(pAcc, accCor);

	float trgVelAbs = sqrtf(pVel[0]*pVel[0] + pVel[1]*pVel[1]);
	float trgAccAbs = sqrtf(pAcc[0]*pAcc[0] + pAcc[1]*pAcc[1]);

	float cruiseCor[2];
	cruiseCor[0] = velCor[0]*trgVelAbs;
	cruiseCor[1] = velCor[1]*trgVelAbs;

	if(trgVelAbs < 0.01f || trgAccAbs < 0.01f)
	{
		memcpy(pCorr, cruiseCor, sizeof(float)*2);
		return;
	}

	float velAng = atan2f(pVel[1], pVel[0]);
	float accAng = atan2f(pAcc[1], pAcc[0]);
	float diff = fabsf(AngleNormalize(velAng-accAng));

	float cor[2];
	if(diff > piHalf)
	{
		// decelerating
		float decelCor[2];
		decelCor[0] = -velCor[0]*trgVelAbs;
		decelCor[1] = -velCor[1]*trgVelAbs;

		if(diff > piHalf+cruiseWindow)
		{
			// out of blending zone
			memcpy(cor, decelCor, sizeof(float)*2);
		}
		else
		{
			// blend with cruiseCor
			float a = (piHalf+cruiseWindow-diff)/cruiseWindow;
			cor[0] = a*cruiseCor[0] + (1.0f-a)*decelCor[0];
			cor[1] = a*cruiseCor[1] + (1.0f-a)*decelCor[1];
		}
	}
	else
	{
		// accelerating
		float accelCor[2];
		accelCor[0] = velCor[0]*trgVelAbs + accCor[0]*trgAccAbs*trgVelAbs;
		accelCor[1] = velCor[1]*trgVelAbs + accCor[1]*trgAccAbs*trgVelAbs;

		if(diff < piHalf-cruiseWindow)
		{
			// out of blending zone
			memcpy(cor, accelCor, sizeof(float)*2);
		}
		else
		{
			// blend with cruiseCor
			float a = (diff-(piHalf-cruiseWindow))/cruiseWindow;
			cor[0] = a*cruiseCor[0] + (1.0f-a)*accelCor[0];
			cor[1] = a*cruiseCor[1] + (1.0f-a)*accelCor[1];
		}
	}

	memcpy(pCorr, cor, sizeof(float)*2);
}


static inline void velFeedforwardCorrection(const float* pVel, const float* pAcc, float* pOutVel)
{
	float cor[2];

	velCorrection(pVel, pAcc, cor);

	pOutVel[0] = pVel[0] + EMAFilterUpdate(&ctrlTigga.emaCorrection[0], cor[0]);
	pOutVel[1] = pVel[1] + EMAFilterUpdate(&ctrlTigga.emaCorrection[1], cor[1]);
}

static inline void velFeedbackCorrection(const float* pVel, const float* pAcc, float* pOutVel)
{
	float cor[2];

	velCorrection(pVel, pAcc, cor);

	pOutVel[0] = pVel[0] - cor[0];
	pOutVel[1] = pVel[1] - cor[1];
}


/*
                 v, a    +------------+
                         |   Model    |
                  +----> | Inversion  | +----+
                  |      |            |      |
                  |      +------------+      |
                  |                          |
+-----------+     |      +------------+   v  |     +-------+
|Trajectory | +---+      |            |      |     |       |
|Generator  |            | PD Control | +----+---> | Plant |
|           | +--------> |            |            |       |
+-----------+            +------------+            +-------+
                 p
                                ^           p          +
                                |                      |
                                +----------------------+
 */

static void trajGen(const DriveInput* pDrive, RobotCtrlReference* pRef)
{
	const float dT = CTRL_DELTA_T;

	switch(pDrive->modeXY)
	{
		case DRIVE_MODE_GLOBAL_POS:
		{
			TrajGeneratorCreateGlobalPosXY(&ctrlTigga.trajGen, pDrive->pos);

			float lastAccRef[2];
			memcpy(lastAccRef, pRef->trajAccLocal, sizeof(float)*2);

			const float* trgPos = ctrlTigga.trajGen.trajGlobalPos;
			const float* trgVel = ctrlTigga.trajGen.trajGlobalVel;
			const float* trgAcc = ctrlTigga.trajGen.trajGlobalAcc;

			memcpy(pRef->trajPos, trgPos, sizeof(float)*2);

			CtrlUtilTurnGlobal2Local(trgPos[2], trgVel[0], trgVel[1], pRef->trajVelLocal, pRef->trajVelLocal+1);
			CtrlUtilTurnGlobal2Local(trgPos[2], trgAcc[0], trgAcc[1], pRef->trajAccLocal, pRef->trajAccLocal+1);

			pRef->trajJerkLocal[0] = (pRef->trajAccLocal[0]-lastAccRef[0])/dT;
			pRef->trajJerkLocal[1] = (pRef->trajAccLocal[1]-lastAccRef[1])/dT;

			TrajGeneratorStepXY(&ctrlTigga.trajGen, dT);
		}
		break;
		case DRIVE_MODE_LOCAL_VEL:
		{
			TrajGeneratorCreateLocalVelXY(&ctrlTigga.trajGen, pDrive->localVel);

			memcpy(pRef->trajPos, ctrlTigga.trajGen.filterGlobalPos, sizeof(float)*2);
			memcpy(pRef->trajVelLocal, ctrlTigga.trajGen.trajLocalVel, sizeof(float)*2);
			memcpy(pRef->trajAccLocal, ctrlTigga.trajGen.trajLocalAcc, sizeof(float)*2);
			memcpy(pRef->trajJerkLocal, ctrlTigga.trajGen.trajLocalJerk, sizeof(float)*2);

			TrajGeneratorStepXY(&ctrlTigga.trajGen, dT);
		}
		break;
		default:
		{
			TrajGeneratorResetXY(&ctrlTigga.trajGen);
		}
		break;
	}

	switch(pDrive->modeW)
	{
		case DRIVE_MODE_GLOBAL_POS:
		{
			TrajGeneratorCreateGlobalPosW(&ctrlTigga.trajGen, pDrive->pos[2]);

			float lastAccRef = pRef->trajAccLocal[2];

			pRef->trajPos[2] = ctrlTigga.trajGen.trajGlobalPos[2];
			pRef->trajVelLocal[2] = ctrlTigga.trajGen.trajGlobalVel[2];
			pRef->trajAccLocal[2] = ctrlTigga.trajGen.trajGlobalAcc[2];
			pRef->trajJerkLocal[2] = (pRef->trajAccLocal[2]-lastAccRef)/dT;

			TrajGeneratorStepW(&ctrlTigga.trajGen, dT);
		}
		break;
		case DRIVE_MODE_LOCAL_VEL:
		{
			// Trajectory Update
			TrajGeneratorCreateLocalVelW(&ctrlTigga.trajGen, pDrive->localVel[2]);

			pRef->trajPos[2] = ctrlTigga.trajGen.filterGlobalPos[2];
			pRef->trajVelLocal[2] = ctrlTigga.trajGen.trajLocalVel[2];
			pRef->trajAccLocal[2] = ctrlTigga.trajGen.trajLocalAcc[2];
			pRef->trajJerkLocal[2] = ctrlTigga.trajGen.trajLocalJerk[2];

			TrajGeneratorStepW(&ctrlTigga.trajGen, dT);
		}
		break;
		default:
		{
			TrajGeneratorResetW(&ctrlTigga.trajGen);
		}
		break;
	}
}

static void controlXY(const DriveInput* pDrive, const RobotCtrlReference* pReference,
		const RobotCtrlState* pState, uint8_t ctrlFirstRun, float* pLocalOutVel, float* pLocalOutAcc)
{
	const float dT = CTRL_DELTA_T;

	switch(pDrive->modeXY)
	{
		case DRIVE_MODE_GLOBAL_POS:
		{
			float trgVelLocal[2];
			float trgAccLocal[2];

			// use current state to rotate traj output
			CtrlUtilRotate(pReference->trajPos[2]-pState->pos[2], pReference->trajVelLocal[0], pReference->trajVelLocal[1], trgVelLocal);
			CtrlUtilRotate(pReference->trajPos[2]-pState->pos[2], pReference->trajAccLocal[0], pReference->trajAccLocal[1], trgAccLocal);

			// XY position control
			float pidOutputXY[2];
			updatePID2D(&ctrlTigga.posPIDXY, pReference->trajPos, pState->pos, pidOutputXY, ctrlFirstRun, FLT_MAX);

			// convert to local pos PID output
			float localPosPidOutput[2];
			CtrlUtilTurnGlobal2Local(pState->pos[2], pidOutputXY[0], pidOutputXY[1], localPosPidOutput, localPosPidOutput+1);

			float vel[2] = {0, 0};
			if(ctrlTigga.configMisc.useFeedforward)
			{
				// feedforward control
				velFeedforwardCorrection(trgVelLocal, trgAccLocal, vel);
			}
			else
			{
				memcpy(vel, trgVelLocal, sizeof(float)*2);
			}

			pLocalOutVel[0] = vel[0]+localPosPidOutput[0];
			pLocalOutVel[1] = vel[1]+localPosPidOutput[1];

			memcpy(pLocalOutAcc, trgAccLocal, sizeof(float)*2);

			// adjust vel/acc if near target
//			float errorPosX = pState->pos[0] - pDrive->pos[0];
//			float errorPosY = pState->pos[1] - pDrive->pos[1];
//			float errorXY = sqrtf(errorPosX*errorPosX+errorPosY*errorPosY);
//			if(errorXY > 0.01f)
//			{
//				memcpy(pLocalOutAcc, trgAccLocal, sizeof(float)*2);
//			}
//			else
//			{
//				float scale = errorXY/0.01f;
//				pLocalOutAcc[0] = trgAccLocal[0]*scale;
//				pLocalOutAcc[1] = trgAccLocal[1]*scale;
//				pLocalOutVel[0] *= scale;
//				pLocalOutVel[1] *= scale;
//			}

			// compensate rotation
			float rot = -pState->vel[2]*dT*0.5f;
			CtrlUtilRotate(rot, pLocalOutVel[0], pLocalOutVel[1], pLocalOutVel);
			CtrlUtilRotate(rot, pLocalOutAcc[0], pLocalOutAcc[1], pLocalOutAcc);
		}
		break;
		case DRIVE_MODE_LOCAL_VEL:
		{
			const float* trgVel = pReference->trajVelLocal;
			const float* trgAcc = pReference->trajAccLocal;

			// XY velocity control
			float pidOutputXY[2];
			updatePID2D(&ctrlTigga.velPIDXY, trgVel, pState->vel, pidOutputXY, ctrlFirstRun, FLT_MAX);

			float vel[2] = {0, 0};
			if(ctrlTigga.configMisc.useFeedforward)
			{
				// feedforward control
				velFeedforwardCorrection(trgVel, trgAcc, vel);
			}
			else
			{
				memcpy(vel, trgVel, sizeof(float)*2);
			}

			pLocalOutVel[0] = vel[0]+pidOutputXY[0];
			pLocalOutVel[1] = vel[1]+pidOutputXY[1];
			memcpy(pLocalOutAcc, trgAcc, sizeof(float)*2);

			// compensate rotation
			float rot = -pState->vel[2]*dT*0.5f;
			CtrlUtilRotate(rot, pLocalOutVel[0], pLocalOutVel[1], pLocalOutVel);
			CtrlUtilRotate(rot, pLocalOutAcc[0], pLocalOutAcc[1], pLocalOutAcc);
		}
		break;
		default:
		{
			ctrlTigga.emaCorrection[0].value = 0;
			ctrlTigga.emaCorrection[1].value = 0;
		}
		break;
	}
}

static void controlW(const DriveInput* pDrive, const RobotCtrlReference* pReference,
		const RobotCtrlState* pState, uint8_t ctrlFirstRun, float* pLocalOutVel, float* pLocalOutAcc)
{
	switch(pDrive->modeW)
	{
		case DRIVE_MODE_GLOBAL_POS:
		{
			// orientation processing
			ctrlTigga.posPIDW.setpoint = pReference->trajPos[2];
			if(ctrlFirstRun)
				ctrlTigga.posPIDW.lastMeasurement = pState->pos[2];
			PIDUpdate(&ctrlTigga.posPIDW, pState->pos[2]);

			// convert to local pos PID output
			float localPosPidOutput = ctrlTigga.posPIDW.output;

			pLocalOutVel[2] = localPosPidOutput + pReference->trajVelLocal[2];

			pLocalOutAcc[2] = pReference->trajAccLocal[2];

			// adjust vel/acc if near target
//			float errW = fabsf(pState->pos[2]-pDrive->pos[2]);
//			if(errW > 0.05f)
//			{
//				pLocalOutAcc[2] = pReference->trajAccLocal[2];
//			}
//			else
//			{
//				float scale = errW/0.05f;
//				pLocalOutAcc[2] = scale*pReference->trajAccLocal[2];
//				pLocalOutVel[2] *= scale;
//			}
		}
		break;
		case DRIVE_MODE_LOCAL_VEL:
		{
			float localVel = pState->vel[2];

			// W velocity control
			ctrlTigga.velPIDW.setpoint = pReference->trajVelLocal[2];
			if(ctrlFirstRun)
				ctrlTigga.velPIDW.lastMeasurement = localVel;
			PIDUpdate(&ctrlTigga.velPIDW, localVel);

			pLocalOutVel[2] = ctrlTigga.velPIDW.output + pReference->trajVelLocal[2];
			pLocalOutAcc[2] = pReference->trajAccLocal[2];
		}
		break;
		default:
		{
		}
		break;
	}
}

static void controller(const RobotCtrlState* pState, const DriveInput* pDrive, RobotCtrlOutput* pOutput, RobotCtrlReference* pReference)
{
	(void)pState;

	uint32_t startTime = SysTimeUSecRaw();

	// Configure trajectory generator
	ctrlTigga.trajGen.vMaxXY = pDrive->limits.velMaxXY;
	ctrlTigga.trajGen.vMaxW = pDrive->limits.velMaxW;
	ctrlTigga.trajGen.aMaxXY = pDrive->limits.accMaxXY;
	ctrlTigga.trajGen.aMaxW = pDrive->limits.accMaxW;
	ctrlTigga.trajGen.jMaxXY = pDrive->limits.jerkMaxXY;
	ctrlTigga.trajGen.jMaxW = pDrive->limits.jerkMaxW;

	// create predicted state for control
	RobotCtrlState ctrlState;
	memcpy(ctrlState.pos, ctrlTigga.mechState, sizeof(float)*3);
	memcpy(ctrlState.vel, ctrlTigga.mechState+3, sizeof(float)*3);
	memcpy(ctrlState.acc, ctrlTigga.mechState+6, sizeof(float)*2);
	ctrlState.acc[2] = 0;

	// create local drive input with multi-turn orientation
	DriveInput drive;
	memcpy(&drive, pDrive, sizeof(DriveInput));
	drive.pos[2] = ctrlState.pos[2] + AngleNormalize(drive.pos[2] - AngleNormalize(ctrlState.pos[2]));

	// set current filter state on traj generator
	TrajGeneratorSetState(&ctrlTigga.trajGen, ctrlState.pos, ctrlState.vel);

	// run trajectory generator
	trajGen(&drive, pReference);

	// determine if controllers do a first run
	uint8_t ctrlFirstRunXY = 0;
	uint8_t ctrlFirstRunW = 0;

	if(pDrive->modeXY != ctrlTigga.lastModeXY)
	{
		ctrlTigga.lastModeXY = pDrive->modeXY;
		ctrlFirstRunXY = 1;
	}

	if(pDrive->modeW != ctrlTigga.lastModeW)
	{
		ctrlTigga.lastModeW = pDrive->modeW;
		ctrlFirstRunW = 1;
	}

	// call controllers for XY and W
	float outVel[3] = {0, 0, 0};
	float outAcc[3] = {0, 0, 0};

	controlXY(&drive, pReference, &ctrlState, ctrlFirstRunXY, outVel, outAcc);
	controlW(&drive, pReference, &ctrlState, ctrlFirstRunW, outVel, outAcc);

	// transform to motor velocities
	CtrlUtilLocalVelToMotorVel(outVel, pOutput->motorVel, ctrlTigga.configMisc.ASR);
	CtrlUtilLocalVelToMotorVel(outAcc, pOutput->motorAcc, ctrlTigga.configMisc.ASR);
	pOutput->ctrlMode = MOT_CTRL_VEL_ACC;

	if(pDrive->modeXY == DRIVE_MODE_WHEEL_VEL || pDrive->modeW == DRIVE_MODE_WHEEL_VEL)
	{
		for(uint8_t i = 0; i < 4; i++)
			pOutput->motorVel[i] += pDrive->wheelVel[i]*CTRL_WHEEL_TO_MOTOR_RATIO;
	}

	if(pDrive->modeXY == DRIVE_MODE_OFF && pDrive->modeW == DRIVE_MODE_OFF)
	{
		memset(pOutput->motorVel, 0, sizeof(float)*4);

		pOutput->ctrlMode = MOT_CTRL_OFF;
	}

	if(chVTTimeElapsedSinceX(ctrlTigga.lastInitTime) < MS2ST(500))
	{
		pOutput->ctrlMode = MOT_CTRL_OFF;
	}
	else
	{
		ctrlTigga.lastInitTime = chVTGetSystemTimeX() - MS2ST(1000);
	}

	if(ctrlTigga.filterUnstable)
	{
		pOutput->ctrlMode = MOT_CTRL_OFF;
	}

	if(pOutput->ctrlMode == MOT_CTRL_OFF)
	{
		++ctrlTigga.motorOffCounter;
	}
	else
	{
		ctrlTigga.motorOffCounter = 0;
	}

	if(ctrlTigga.motorOffCounter > 5000)
	{
		ctrlTigga.motorOffCounter = 0;

		float subtract = ctrlTigga.vision.turns*2.0f*M_PI;
		ctrlTigga.mainState[2] -= subtract;
		ctrlTigga.mechState[2] -= subtract;
		ctrlTigga.ekf.x.pData[2] -= subtract;

		for(uint16_t i = 0; i < ctrlTigga.ins.size; i++)
			ctrlTigga.ins.buffer[i][2] -= subtract;

		ctrlTigga.vision.turns = 0;

		if(ctrlTigga.visLost == 0 && ctrlTigga.filterUnstable == 0)
		{
			ctrlTigga.bias[0] = ctrlTigga.emaBias[0].value;
			ctrlTigga.bias[1] = ctrlTigga.emaBias[1].value;
			ctrlTigga.bias[2] = ctrlTigga.emaBias[2].value;
		}
	}

	ctrlTigga.log.ctrlTime = SysTimeUSecRaw()-startTime;
	ctrlTigga.log.header.timestamp = SysTimeUSec();
}

static void sensorFusion(const RobotSensors* pSensors, const RobotCtrlOutput* pOutput, RobotCtrlState* pState)
{
	uint32_t startTime = SysTimeUSecRaw();

	// VISION
	const float* pLateVis = visionProc(pSensors->vision.updated, pSensors->vision.delay, pSensors->vision.pos);
	if(pSensors->vision.noVision)
	{
		ctrlTigga.filterUnstable = 0;
	}

	// GYRO + ACCELEROMETER
	float gyrAcc[3] = {pSensors->gyr.rotVel[2]-ctrlTigga.bias[0], pSensors->acc.linAcc[0]-ctrlTigga.bias[1], pSensors->acc.linAcc[1]-ctrlTigga.bias[2]};
	memcpy(ctrlTigga.gyrAcc.buffer[ctrlTigga.gyrAcc.writePtr], gyrAcc, sizeof(float)*3);
	ctrlTigga.gyrAcc.writePtr = (ctrlTigga.gyrAcc.writePtr + 1) % ctrlTigga.gyrAcc.size;
	const float* pLateGyrAcc = ctrlTigga.gyrAcc.buffer[ctrlTigga.gyrAcc.writePtr];

	if(pOutput->ctrlMode == MOT_CTRL_OFF && ctrlTigga.visLost == 0)
	{
		EMAFilterUpdate(&ctrlTigga.emaBias[0], pSensors->gyr.rotVel[2]);
		EMAFilterUpdate(&ctrlTigga.emaBias[1], pSensors->acc.linAcc[0]);
		EMAFilterUpdate(&ctrlTigga.emaBias[2], pSensors->acc.linAcc[1]);
	}

	// ENCODERS
	float encVelLocal[3];
	float motAccOutLocal[3];
	// get local velocity from encoders
	CtrlUtilMotorVelToLocalVel(pSensors->enc.vel, encVelLocal);
	// get local acc from previous output
	CtrlUtilMotorVelToLocalVel(pOutput->motorAcc, motAccOutLocal);
	// correct local velocity
	if(ctrlTigga.configMisc.useFeedforward)
		velFeedbackCorrection(encVelLocal, motAccOutLocal, encVelLocal);

	// store local vel from encoders in buffer
	memcpy(ctrlTigga.enc.buffer[ctrlTigga.enc.writePtr], encVelLocal, sizeof(float)*3);
	ctrlTigga.enc.writePtr = (ctrlTigga.enc.writePtr + 1) % ctrlTigga.enc.size;
	const float* pLateEncVelLocal = ctrlTigga.enc.buffer[ctrlTigga.enc.writePtr];

	// NO VISION handling
	++ctrlTigga.noVisionCounter;
	if(pLateVis)
	{
		if(ctrlTigga.visLost)
		{
			// vision restored, check position
			float xPosErr = pLateVis[0] - ctrlTigga.ekf.x.pData[0];
			float yPosErr = pLateVis[1] - ctrlTigga.ekf.x.pData[1];
			float xyPosErr = sqrtf(xPosErr*xPosErr+yPosErr*yPosErr);
			float angErr = fabsf(AngleNormalize(pLateVis[2]-ctrlTigga.ekf.x.pData[2]));
			if(xyPosErr > 0.1f || angErr > 0.1f)
			{
				// large error from EKF, consider filter as unstable
				ctrlTigga.filterUnstable = 1;
			}
		}
		memcpy(ctrlTigga.enc.pos, pLateVis, sizeof(float)*3);
		ctrlTigga.noVisionCounter = 0;
		ctrlTigga.visLost = 0;
	}

	float encVelGlobal[3];
	CtrlUtilTurnLocal2Global(ctrlTigga.mainState[2], pLateEncVelLocal[0], pLateEncVelLocal[1], encVelGlobal, encVelGlobal+1);
//	encVelGlobal[2] = pLateGyrAcc[0];
	encVelGlobal[2] = pLateEncVelLocal[2];

	for(uint8_t i = 0; i < 3; i++)
		ctrlTigga.enc.pos[i] += encVelGlobal[i]*CTRL_DELTA_T;

	if(ctrlTigga.noVisionCounter > ctrlTigga.configFilter.visionTimeoutMs)
	{
		ctrlTigga.visLost = 1;

		// copy filter position to encoder position to avoid sudden jump of position
		memcpy(ctrlTigga.enc.pos, ctrlTigga.mechState, sizeof(float)*3);
	}

	if( ctrlTigga.mainState[0] > ctrlTigga.configMisc.fieldSizeX*0.5f || ctrlTigga.mainState[0] < -ctrlTigga.configMisc.fieldSizeX*0.5f ||
		ctrlTigga.mainState[1] > ctrlTigga.configMisc.fieldSizeY*0.5f || ctrlTigga.mainState[1] < -ctrlTigga.configMisc.fieldSizeY*0.5f)
	{
		ctrlTigga.visLost = 1;
	}

	// INERTIAL NAVIGATION SYSTEM (INS)
	const float* pInsLast = ctrlTigga.ins.buffer[(ctrlTigga.ins.writePtr+ctrlTigga.ins.size-1)%ctrlTigga.ins.size];
	arm_matrix_instance_f32 matInsState = {5, 1, (float[5]){}};
	memcpy(matInsState.pData, pInsLast, sizeof(float)*5);
	const arm_matrix_instance_f32 matInput = {3, 1, gyrAcc};
	(*ctrlTigga.ekf.pState)(&matInsState, &matInput);

	memcpy(ctrlTigga.ins.buffer[ctrlTigga.ins.writePtr], matInsState.pData, sizeof(float)*5);
	ctrlTigga.ins.writePtr = (ctrlTigga.ins.writePtr + 1) % ctrlTigga.ins.size;
	const float* pLateIns = ctrlTigga.ins.buffer[ctrlTigga.ins.writePtr];

	// EKF
	memcpy(ctrlTigga.ekf.u.pData, pLateGyrAcc, sizeof(float)*3);
	EKFPredict(&ctrlTigga.ekf);

	if(pLateVis)
	{
		memcpy(ctrlTigga.ekf.z.pData, pLateVis, sizeof(float)*3);
		EKFUpdate(&ctrlTigga.ekf);

		float xPosErr = pLateVis[0] - ctrlTigga.ekf.x.pData[0];
		float yPosErr = pLateVis[1] - ctrlTigga.ekf.x.pData[1];
		float xyPosErr = sqrtf(xPosErr*xPosErr+yPosErr*yPosErr);
		float angErr = fabsf(AngleNormalize(pLateVis[2]-ctrlTigga.ekf.x.pData[2]));
		if(xyPosErr > 0.1f || angErr > 0.1f)
		{
			// large position error
		}
		else
		{
			// small error, consider filter stable
			if(ctrlTigga.filterUnstable)
				TrajGeneratorReset(&ctrlTigga.trajGen);

			ctrlTigga.filterUnstable = 0;
		}
	}
	else
	{
		uint16_t encDt = (uint16_t)(1000.0f/((float)ctrlTigga.configFilter.encPosRate)+0.5f);
		if(ctrlTigga.visLost && ctrlTigga.noVisionCounter >= encDt)
		{
			memcpy(ctrlTigga.ekf.z.pData, ctrlTigga.enc.pos, sizeof(float)*3);
			EKFUpdate(&ctrlTigga.ekf);
			ctrlTigga.noVisionCounter = 0;
		}
	}

	// TRACKING (INS follows EKF)
	float gain = CTRL_DELTA_T/(ctrlTigga.configFilter.trackingCoeff*ctrlTigga.ins.size*0.001f);
	float corr[5];
	for(uint8_t i = 0; i < 5; i++)
		corr[i] = (ctrlTigga.ekf.x.pData[i]-pLateIns[i])*gain;

	for(uint16_t i = 0; i < ctrlTigga.ins.size; i++)
	{
		for(uint16_t j = 0; j < 5; j++)
			ctrlTigga.ins.buffer[i][j] += corr[j];
	}

	// GATHER STATES
	memcpy(ctrlTigga.mainState, ctrlTigga.ekf.x.pData, sizeof(float)*5);
	ctrlTigga.mainState[5] = pLateGyrAcc[0];
	ctrlTigga.mainState[6] = pLateGyrAcc[1];
	ctrlTigga.mainState[7] = pLateGyrAcc[2];

	pInsLast = ctrlTigga.ins.buffer[(ctrlTigga.ins.writePtr+ctrlTigga.ins.size-1)%ctrlTigga.ins.size];
	memcpy(ctrlTigga.mechState, pInsLast, sizeof(float)*5);
	ctrlTigga.mechState[5] = gyrAcc[0];
	ctrlTigga.mechState[6] = gyrAcc[1];
	ctrlTigga.mechState[7] = gyrAcc[2];

	// FILL GLOBAL STATE
	memcpy(pState->pos, ctrlTigga.mechState, sizeof(float)*3);
	pState->pos[2] = AngleNormalize(ctrlTigga.mechState[2]);

	float globalVel[3] = {0, 0, ctrlTigga.mechState[5]};
	CtrlUtilTurnLocal2Global(ctrlTigga.mechState[2], ctrlTigga.mechState[3], ctrlTigga.mechState[4], globalVel, globalVel+1);
	memcpy(pState->vel, globalVel, sizeof(float)*3);

	float globalAcc[2];
	CtrlUtilTurnLocal2Global(ctrlTigga.mechState[2], ctrlTigga.mechState[6], ctrlTigga.mechState[7], globalAcc, globalAcc+1);
	pState->acc[0] = globalAcc[0];
	pState->acc[1] = globalAcc[1];
	pState->acc[2] = 0;	// can't estimate that with this filter

	pState->posUpdated = 1;
	pState->velUpdated = 1;
	pState->accUpdated = 1;

	ctrlTigga.log.fusionTime = SysTimeUSecRaw()-startTime;
}

void CtrlTiggaInitCfg()
{
	ctrlTigga.pConfigFileFilter = ConfigOpenOrCreate(&configFileDescFilter, &ctrlTigga.configFilter, sizeof(CtrlTiggaConfigFilter), &configUpdateCallback, 0);
	ctrlTigga.pConfigFilePid = ConfigOpenOrCreate(&configFileDescPid, &ctrlTigga.configPid, sizeof(CtrlTiggaConfigPID), &configUpdateCallback, 0);
	ctrlTigga.pConfigFileMisc = ConfigOpenOrCreate(&configFileDescMisc, &ctrlTigga.configMisc, sizeof(CtrlTiggaConfigMisc), &configUpdateCallback, 0);
	ctrlTigga.pConfigBotModel = ConfigOpenOrCreate(&configFileDescBotModel, &ctrlTigga.configBotModel, sizeof(CtrlTiggaConfigBotModel), &configUpdateCallback, 0);
}

static inline const float* visionProc(uint8_t updated, int32_t tDelayUs, const float* pVisPos)
{
	float* pRet = 0;

	// store maximum encountered vision delay
	if(tDelayUs > ctrlTigga.maxVisionDelay)
		ctrlTigga.maxVisionDelay = tDelayUs;

	// check if a sample is in the buffer for the current time
	if(ctrlTigga.vision.buffer[ctrlTigga.vision.readPtr].state == 1)
	{
		pRet = ctrlTigga.vision.buffer[ctrlTigga.vision.readPtr].pos;
	}

	// slot consumed
	ctrlTigga.vision.buffer[ctrlTigga.vision.readPtr].state = 0;

	// advance read pointer
	ctrlTigga.vision.readPtr = (ctrlTigga.vision.readPtr + 1)%ctrlTigga.vision.size;

	if(updated)
	{
		tDelayUs += ((uint32_t)ctrlTigga.configFilter.visCaptureDelay)*1000;

		// validate sample, outlier rejection
		uint32_t measTime = SysTimeUSecRaw()-tDelayUs;
		float visionDt = (measTime-ctrlTigga.lastVisionTime)*1e-6f;

		float searchRadiusXY = ctrlTigga.configFilter.outlierMaxVelXY*visionDt;
		float searchRadiusW = ctrlTigga.configFilter.outlierMaxVelW*visionDt;

		float diffX = pVisPos[0] - ctrlTigga.vision.lastVisionPos[0];
		float diffY = pVisPos[1] - ctrlTigga.vision.lastVisionPos[1];
		float diffXY = sqrtf(diffX*diffX+diffY*diffY);
		float diffW = AngleNormalize(pVisPos[2] - ctrlTigga.vision.lastVisionPos[2]);

		if(diffXY > searchRadiusXY || fabsf(diffW) > searchRadiusW)
		{
			// this is an invalid sample, vel is impossible
			return pRet;
		}

		ctrlTigga.lastVisionTime = measTime;
		memcpy(ctrlTigga.vision.lastVisionPos, pVisPos, sizeof(float)*3);

		// multi-turn angle correction
		if(pVisPos[2] < -2.0f && ctrlTigga.vision.lastVisionOrient > 2.0f)
			++ctrlTigga.vision.turns;

		if(pVisPos[2] > 2.0f && ctrlTigga.vision.lastVisionOrient < -2.0f)
			--ctrlTigga.vision.turns;

		ctrlTigga.vision.lastVisionOrient = pVisPos[2];

		float orient = pVisPos[2] + ctrlTigga.vision.turns*2.0f*M_PI;

		// calculate vision buffer insertion point
		int32_t visTime = ctrlTigga.vision.size*CTRL_DELTA_T_IN_US;
		int32_t insertPos = (visTime-tDelayUs+CTRL_DELTA_T_IN_US/2)/CTRL_DELTA_T_IN_US - 1;
		if(insertPos <= 0)
		{
			// this sample is too late to be inserted at the correct location, log the error...
			++ctrlTigga.numLateVisionMeasurements;

			// ... and use it anyway with adjusted insertion pos. That's better than to discard it completely.
			insertPos = 1;
		}

		insertPos += ctrlTigga.vision.readPtr;
		insertPos %= ctrlTigga.vision.size;

		// store position
		memcpy(ctrlTigga.vision.buffer[insertPos].pos, pVisPos, sizeof(float)*2);
		ctrlTigga.vision.buffer[insertPos].pos[2] = orient;
		ctrlTigga.vision.buffer[insertPos].state = 1;
	}

	return pRet;
}

static void ekfStateFunc(arm_matrix_instance_f32* pX, const arm_matrix_instance_f32* pU)
{
	const float dt = 0.001f;

	float gyr_w = MAT_ELEMENT(*pU, 0, 0);
	float acc_x = MAT_ELEMENT(*pU, 1, 0);
	float acc_y = MAT_ELEMENT(*pU, 2, 0);

	float p_x = MAT_ELEMENT(*pX, 0, 0);
	float p_y = MAT_ELEMENT(*pX, 1, 0);
	float p_w = MAT_ELEMENT(*pX, 2, 0);
	float v_x = MAT_ELEMENT(*pX, 3, 0);
	float v_y = MAT_ELEMENT(*pX, 4, 0);

	float v_w = gyr_w;
	float a = -M_PI_2 + p_w + dt*v_w*0.5f;
	float a_x = acc_x + v_y*v_w;
	float a_y = acc_y - v_x*v_w;

	float px1 = p_x + (arm_cos_f32(a)*v_x-arm_sin_f32(a)*v_y)*dt + (arm_cos_f32(a)*a_x-arm_sin_f32(a)*a_y)*0.5f*dt*dt;
	float py1 = p_y + (arm_sin_f32(a)*v_x+arm_cos_f32(a)*v_y)*dt + (arm_sin_f32(a)*a_x+arm_cos_f32(a)*a_y)*0.5f*dt*dt;
	float vx1 = v_x + a_x*dt;
	float vy1 = v_y + a_y*dt;
	float pw1 = p_w + v_w*dt;

	MAT_ELEMENT(*pX, 0, 0) = px1;
	MAT_ELEMENT(*pX, 1, 0) = py1;
	MAT_ELEMENT(*pX, 2, 0) = pw1;
	MAT_ELEMENT(*pX, 3, 0) = vx1;
	MAT_ELEMENT(*pX, 4, 0) = vy1;
}

static void ekfStateJacobianFunc(const arm_matrix_instance_f32* pX, const arm_matrix_instance_f32* pU, arm_matrix_instance_f32* pF)
{
	float gyr_w = MAT_ELEMENT(*pU, 0, 0);
	float acc_x = MAT_ELEMENT(*pU, 1, 0);
	float acc_y = MAT_ELEMENT(*pU, 2, 0);

	float p_w = MAT_ELEMENT(*pX, 2, 0);
	float v_x = MAT_ELEMENT(*pX, 3, 0);
	float v_y = MAT_ELEMENT(*pX, 4, 0);

    float SF1 = gyr_w/2000.0f + p_w - M_PI_2;
	float SF2 = arm_cos_f32(SF1)/1000.0f + (gyr_w * arm_sin_f32(SF1))/2000000.0f;
	float SF3 = (gyr_w * arm_cos_f32(SF1))/2000000.0f;
	float SF4 = arm_sin_f32(SF1);
	float SF5 = acc_y - gyr_w * v_x;
	float SF6 = arm_cos_f32(SF1);
	float SF7 = gyr_w/1000.0f;

	arm_mat_identity_f32(pF);

	MAT_ELEMENT(*pF, 0, 2) = - (SF5*SF6)/2000000.0f - (v_x*SF4)/1000.0f - (v_y*SF6)/1000.0f - (SF4*(acc_x + gyr_w*v_y))/2000000.0f;
	MAT_ELEMENT(*pF, 0, 3) = SF2;
	MAT_ELEMENT(*pF, 0, 4) = SF3 - SF4/1000.0f;

	MAT_ELEMENT(*pF, 1, 2) = (v_x*SF6)/1000.0f - (SF4*SF5)/2000000.0f - (v_y*SF4)/1000.0f + (SF6*(acc_x + gyr_w*v_y))/2000000.0f;
	MAT_ELEMENT(*pF, 1, 3) = SF4/1000.0f - SF3;
	MAT_ELEMENT(*pF, 1, 4) = SF2;

	MAT_ELEMENT(*pF, 3, 4) = SF7;

	MAT_ELEMENT(*pF, 4, 3) = -SF7;
}

static void ekfMeasFunc(const arm_matrix_instance_f32* pX, arm_matrix_instance_f32* pY)
{
	memcpy(pY->pData, pX->pData, sizeof(float)*3);
}

static void ekfMeasJacobianFunc(const arm_matrix_instance_f32* pX, arm_matrix_instance_f32* pH)
{
	(void)pX;

	arm_mat_zero_f32(pH);
	MAT_ELEMENT(*pH, 0, 0) = 1.0f;
	MAT_ELEMENT(*pH, 1, 1) = 1.0f;
	MAT_ELEMENT(*pH, 2, 2) = 1.0f;
}
