/*

 * ctrl_tigga.h
 *
 *  Created on: 26.02.2015
 *      Author: AndreR
 */

// TIGGA is pretty clever, he can control a robot :)

#ifndef CTRL_TIGGA_H_
#define CTRL_TIGGA_H_

#include <arm_math.h>
#include <main/ctrl.h>
#include <stdint.h>
#include "util/config.h"
#include "util/ema_filter.h"
#include "util/trajectory.h"
#include "util/traj_generator.h"
#include "util/pid.h"
#include "util/lag_element.h"
#include "util/ekf.h"

#pragma pack(push, 1)
typedef struct _CtrlTiggaConfigFilter
{
	float posNoiseXY;
	float posNoiseW;
	float velNoiseXY;
	float accBiasNoiseXY;
	float gyrBiasNoiseW;

	float visNoiseXY;
	float visNoiseW;

	float outlierMaxVelXY;
	float outlierMaxVelW;

	float trackingCoeff;

	uint8_t visCaptureDelay;
	uint8_t visProcessingDelayMax;

	uint16_t visionTimeoutMs;
	uint16_t encPosRate;
} CtrlTiggaConfigFilter;

typedef struct _CtrlTiggaConfigMisc
{
	float fieldSizeX;
	float fieldSizeY;
	uint8_t ASR;
	uint8_t useFeedforward;
} CtrlTiggaConfigMisc;

typedef struct _CtrlTiggaConfigPID
{
	struct
	{
		float kp;
		float ki;
		float kd;
	} posXY;

	struct
	{
		float kp;
		float ki;
		float kd;
	} posW;

	struct
	{
		float kp;
		float ki;
		float kd;
	} velXY;

	struct
	{
		float kp;
		float ki;
		float kd;
	} velW;
} CtrlTiggaConfigPID;

typedef struct _CtrlTiggaConfigBotModel
{
	float X[3];
	float Y[3];
} CtrlTiggaConfigBotModel;
#pragma pack(pop)

typedef struct _CtrlTiggaVisionEntry
{
	float pos[3];
	uint8_t state;
} CtrlTiggaVisionEntry;

#define CTRL_TIGGA_MAX_BUFFER_SIZE 200

typedef struct _CtrlTigga
{
	// state estimation stuff
	float mainState[8];
	float mechState[8];

	EKF ekf;
	float ekfData[EKF_DATA_SIZE(8, 3, 3, 8)];

	// buffers
	struct _vision
	{
		CtrlTiggaVisionEntry buffer[CTRL_TIGGA_MAX_BUFFER_SIZE];
		uint16_t readPtr;
		uint16_t size;
		float lastVisionPos[3];
		float lastVisionOrient;
		int32_t turns;
	} vision;

	struct _gyrAcc
	{
		float buffer[CTRL_TIGGA_MAX_BUFFER_SIZE][3]; // [gyr, acc_x, acc_y]
		uint16_t size;
		uint16_t writePtr;
	} gyrAcc;

	struct _enc
	{
		float buffer[CTRL_TIGGA_MAX_BUFFER_SIZE][3];
		uint16_t size;
		uint16_t writePtr;
		float pos[3];
	} enc;

	struct _ins
	{
		float buffer[CTRL_TIGGA_MAX_BUFFER_SIZE][5];
		uint16_t size;
		uint16_t writePtr;
	} ins;

	// control stuff
	uint8_t lastModeXY;
	uint8_t lastModeW;

	TrajGenerator trajGen;

	uint8_t visLost;
	uint32_t lastVisionTime;

	uint8_t filterUnstable;

	uint16_t noVisionCounter;

	uint16_t motorOffCounter;

	uint32_t lastInitTime;

	uint32_t numLateVisionMeasurements;
	uint32_t maxVisionDelay;

	PID posPIDXY;
	PID posPIDW;

	PID velPIDXY;
	PID velPIDW;

	EMAFilter emaCorrection[2];

	EMAFilter emaBias[3];
	float bias[3];

	CtrlTiggaConfigFilter configFilter;
	ConfigFile* pConfigFileFilter;

	CtrlTiggaConfigPID configPid;
	ConfigFile* pConfigFilePid;

	CtrlTiggaConfigMisc configMisc;
	ConfigFile* pConfigFileMisc;

	CtrlTiggaConfigBotModel configBotModel;
	ConfigFile* pConfigBotModel;

	LogCtrlTiggaDebug log;
} CtrlTigga;

extern CtrlTigga ctrlTigga;
extern CtrlInstance ctrlTiggaInstance;

void CtrlTiggaInitCfg();

#endif /* CTRL_TIGGA_H_ */
