/*
 * traj_generator.c
 *
 *  Created on: 26.05.2016
 *      Author: AndreR
 */

#include "traj_generator.h"
#include "main/util.h"
#include <string.h>
#include <math.h>

#define MODE_RESET					0
#define MODE_LOCAL_VEL				1
#define MODE_GLOBAL_POS				2

static void changeModeXY(TrajGenerator* pGen, uint8_t newMode)
{
	if(pGen->modeXY == newMode)
		return;

	switch(pGen->modeXY)
	{
		case MODE_RESET:
		{
			// out of reset, we only have the filtered data available
			memcpy(pGen->trajGlobalPos, pGen->filterGlobalPos, sizeof(float)*2);
			memcpy(pGen->trajGlobalVel, pGen->filterGlobalVel, sizeof(float)*2);
			memset(pGen->trajGlobalAcc, 0, sizeof(float)*2);

			CtrlUtilTurnGlobal2Local(pGen->filterGlobalPos[2], pGen->filterGlobalVel[0], pGen->filterGlobalVel[1], pGen->trajLocalVel, pGen->trajLocalVel+1);
			memset(pGen->trajLocalAcc, 0, sizeof(float)*2);
		}
		break;
		case MODE_GLOBAL_POS:
		{
			switch(newMode)
			{
				case MODE_LOCAL_VEL:
				{
					CtrlUtilTurnGlobal2Local(pGen->trajGlobalPos[2], pGen->trajGlobalVel[0], pGen->trajGlobalVel[1], pGen->trajLocalVel, pGen->trajLocalVel+1);
					CtrlUtilTurnGlobal2Local(pGen->trajGlobalPos[2], pGen->trajGlobalAcc[0], pGen->trajGlobalAcc[1], pGen->trajLocalAcc, pGen->trajLocalAcc+1);
				}
				break;
			}
		}
		break;
		case MODE_LOCAL_VEL:
		{
			switch(newMode)
			{
				case MODE_GLOBAL_POS:
				{
					// position is not stored in vel mode, take it from filter
					memcpy(pGen->trajGlobalPos, pGen->filterGlobalPos, sizeof(float)*2);

					CtrlUtilTurnLocal2Global(pGen->filterGlobalPos[2], pGen->trajLocalVel[0], pGen->trajLocalVel[1], pGen->trajGlobalVel, pGen->trajGlobalVel+1);
					CtrlUtilTurnLocal2Global(pGen->filterGlobalPos[2], pGen->trajLocalAcc[0], pGen->trajLocalAcc[1], pGen->trajGlobalAcc, pGen->trajGlobalAcc+1);
				}
				break;
			}
		}
		break;
	}

	pGen->modeXY = newMode;
}

static void changeModeW(TrajGenerator* pGen, uint8_t newMode)
{
	if(pGen->modeW == newMode)
		return;

	switch(pGen->modeW)
	{
		case MODE_RESET:
		{
			// out of reset, we only have the filtered data available
			pGen->trajGlobalPos[2] = pGen->filterGlobalPos[2];
			pGen->trajGlobalVel[2] = pGen->filterGlobalVel[2];
			pGen->trajGlobalAcc[2] = 0;

			pGen->trajLocalVel[2] = pGen->filterGlobalVel[2];
			pGen->trajLocalAcc[2] = 0;
		}
		break;
		case MODE_GLOBAL_POS:
		{
			switch(newMode)
			{
				case MODE_LOCAL_VEL:
				{
					pGen->trajLocalVel[2] = pGen->trajGlobalVel[2];
					pGen->trajLocalAcc[2] = pGen->trajGlobalAcc[2];
				}
				break;
			}
		}
		break;
		case MODE_LOCAL_VEL:
		{
			switch(newMode)
			{
				case MODE_GLOBAL_POS:
				{
					// position is not stored in vel mode, take it from filter
					pGen->trajGlobalPos[2] = pGen->filterGlobalPos[2];
					pGen->trajGlobalVel[2] = pGen->trajLocalVel[2];
					pGen->trajGlobalAcc[2] = pGen->trajLocalAcc[2];
				}
				break;
			}
		}
		break;
	}

	pGen->modeW = newMode;
}

void TrajGeneratorInit(TrajGenerator* pGen)
{
	memset(pGen, 0, sizeof(*pGen));

	pGen->modeXY = MODE_RESET;
	pGen->modeW = MODE_RESET;

	pGen->vMaxXY = 2;
	pGen->aMaxXY = 3;
	pGen->jMaxXY = 5;
	pGen->vMaxW = 10;
	pGen->aMaxW = 20;
	pGen->jMaxW = 40;
}

void TrajGeneratorReset(TrajGenerator* pGen)
{
	pGen->modeXY = MODE_RESET;
	pGen->modeW = MODE_RESET;
}

void TrajGeneratorResetXY(TrajGenerator* pGen)
{
	pGen->modeXY = MODE_RESET;
}

void TrajGeneratorResetW(TrajGenerator* pGen)
{
	pGen->modeW = MODE_RESET;
}

void TrajGeneratorSetState(TrajGenerator* pGen, const float* pGlobalPos, const float* pGlobalVel)
{
	memcpy(pGen->filterGlobalPos, pGlobalPos, sizeof(float)*3);
	memcpy(pGen->filterGlobalVel, pGlobalVel, sizeof(float)*3);
}

void TrajGeneratorCreateGlobalPosXY(TrajGenerator* pGen, const float* pPosTarget)
{
	changeModeXY(pGen, MODE_GLOBAL_POS);

	TrajSecOrder2DCreate(&pGen->trajPosXY, pGen->trajGlobalPos, pGen->trajGlobalVel, pPosTarget, pGen->vMaxXY, pGen->aMaxXY);
}

void TrajGeneratorCreateGlobalPosW(TrajGenerator* pGen, float orientTarget)
{
	changeModeW(pGen, MODE_GLOBAL_POS);

	TrajSecOrder1DCreate(&pGen->trajPosW, pGen->trajGlobalPos[2], pGen->trajGlobalVel[2], orientTarget, pGen->vMaxW, pGen->aMaxW);
}

void TrajGeneratorCreateLocalVelXY(TrajGenerator* pGen, const float* pVelTarget)
{
	changeModeXY(pGen, MODE_LOCAL_VEL);

	TrajSecOrder2DCreate(&pGen->trajVelXY, pGen->trajLocalVel, pGen->trajLocalAcc, pVelTarget, pGen->aMaxXY, pGen->jMaxXY);
}

void TrajGeneratorCreateLocalVelW(TrajGenerator* pGen, float velTargetW)
{
	changeModeW(pGen, MODE_LOCAL_VEL);

	TrajSecOrder1DCreate(&pGen->trajVelW, pGen->trajLocalVel[2], pGen->trajLocalAcc[2], velTargetW, pGen->aMaxW, pGen->jMaxW);
}

void TrajGeneratorStepXY(TrajGenerator* pGen, float tStep)
{
	switch(pGen->modeXY)
	{
		case MODE_GLOBAL_POS:
		{
			TrajSecOrder2DValuesAtTime(&pGen->trajPosXY, tStep, pGen->trajGlobalPos, pGen->trajGlobalVel, pGen->trajGlobalAcc);
		}
		break;
		case MODE_LOCAL_VEL:
		{
			TrajSecOrder2DValuesAtTime(&pGen->trajVelXY, tStep, pGen->trajLocalVel, pGen->trajLocalAcc, pGen->trajLocalJerk);
		}
		break;
	}
}

void TrajGeneratorStepW(TrajGenerator* pGen, float tStep)
{
	switch(pGen->modeW)
	{
		case MODE_GLOBAL_POS:
		{
			TrajSecOrder1DValuesAtTime(&pGen->trajPosW, tStep, pGen->trajGlobalPos+2, pGen->trajGlobalVel+2, pGen->trajGlobalAcc+2);
		}
		break;
		case MODE_LOCAL_VEL:
		{
			TrajSecOrder1DValuesAtTime(&pGen->trajVelW, tStep, pGen->trajLocalVel+2, pGen->trajLocalAcc+2, pGen->trajLocalJerk+2);
		}
		break;
	}
}
